#Plot control file. Currently used by batch scripts in examples/parallel.
#Place this in snudda/plotting at a later stage
import sys
from snudda.plotting import PlotSpikeRaster
from snudda.plotting import SnuddaPlotSpikeRaster2

from snudda.plotting import PlotTraces
import matplotlib.pyplot as plt
import os

def plot_network_simulation(nametag,doHistRast=True,doTraces=True,doInpRast=True,inputFileName="input-spikes.hdf5",input_type='',neuronType='dSPN'):
    print('Starting plotting network simulation')
    print(nametag)
    print("..............")
    networkName=nametag
    network_path = os.path.join("networks", networkName)
    
    ##############################################################
    if doHistRast:#tes2
        print("Starting SnuddaPlotSpikeRaster2")
        type_order = ["chin", "dspn", "lts", "ispn", "fs", "fsn"]
        type_division = [[ "dspn", "ispn"],["chin", "fs", "fsn", "lts"]]
        sp = SnuddaPlotSpikeRaster2(network_path=network_path)
        #sp.plot_hist_raster(type_order=type_order, fig_size=(5, 10))
        #sp.plot_hist_raster(type_order=type_order,type_division=type_division, fig_size=(5, 10))
        sp.plot_hist_raster(type_division=type_division)   
    
    ##############################################################
    if doTraces:
        print("Starting plot_traces")
        network_file = os.path.join(network_path,"network-synapses.hdf5")
        network_output = os.path.join(network_path,"simulation", "output.hdf5")
        pt=PlotTraces(network_output, network_file=network_file)
        pt.plot_traces_sep(folder_name=input_type)
        pt.plot_traces(fig_name="traces.pdf")
    
    ##############################################################
    if doInpRast:
        print("Starting plot_input")
        from snudda.plotting import PlotInput
        input_file = os.path.join(network_path, inputFileName)
        print("input_file:")
        print(input_file)
        spi = PlotInput(input_file)
        spi.plot_input(neuronType, 10, fig_size=(15, 5))
        plt.xlim([0,10])
        plt.savefig(os.path.join(network_path,"figures",f"inputspikes_{input_type}"), dpi=300)
        print("asd")
    
    ##############################################################
if __name__ == '__main__':
    from argparse import ArgumentParser, RawTextHelpFormatter
    parser = ArgumentParser("Plot network", formatter_class=RawTextHelpFormatter)
    parser.add_argument("--nametag")
    parser.add_argument("--doHistRast",type=int, default=1)
    parser.add_argument("--doTraces",type=int, default=1)
    parser.add_argument("--doInpRast",type=int, default=1)
    parser.add_argument("--inputFileName",type=str, default="input-spikes.hdf5")
    parser.add_argument("--input_type",type=str, default='')
    parser.add_argument("--neuronType",type=str, default='dSPN')
    args = parser.parse_args()
    plot_network_simulation(args.nametag,doHistRast=args.doHistRast,doTraces=args.doTraces,doInpRast=args.doInpRast,inputFileName=args.inputFileName,input_type=args.input_type,neuronType=args.neuronType)




import json
from collections import OrderedDict

import numpy as np
from scipy.interpolate import griddata

from snudda.utils import snudda_parse_path


class ProjectionDetection:

    def __init__(self, snudda_detect, rc=None, role=None):

        """

            Args:
                snudda_detect : SnuddaDetect object
                rc (default None): Ipyparallel remote client

        """

        self.snudda_detect = snudda_detect
        self.rc = rc
        self.d_view = None
        self.workers_initialised = False

        self.hyper_voxel_projections = dict()
        self.projections = dict()

        if not role:
            self.role = "master"
        else:
            self.role = role

        self.parse_config()

        if self.rc is not None:
            self.d_view = self.rc.direct_view(targets='all')
            self.setup_parallel()

    def setup_parallel(self):

        if self.d_view is None:
            return

        """ Prepares workers for parallel execution if d_view is not None. """

        assert self.role == "master", \
            "setup_parallel: Should only be called by master node"

        if self.workers_initialised:
            self.write_log("Workers already initialised.")
            return

        with self.d_view.sync_imports():
            from snudda.detect.projection_detection import ProjectionDetection

        cmd_str = "spd = ProjectionDetection(snudda_detect=sd, role='worker')"
        self.d_view.execute(cmd_str, block=True)
        cmd_str2 = "sd.projection_detection = spd"
        self.d_view.execute(cmd_str2, block=True)

        self.write_log(f"Workers set up.")

    def sync_projection_info(self):

        if self.role == "master" and self.d_view is not None:

            self.write_log("Synchronising projection info to workers.")
            for var_name in ["projections", "hyper_voxel_projections"]:
                self.d_view.push({"tmp_val": eval(f"self.{var_name}")}, block=True)
                cmd_str = f"spd.{var_name} = tmp_val"
                self.d_view.execute(cmd_str, block=True)

    def write_log(self, *args, **kvargs):
        # Reuse the log files for snudda_detect
        self.snudda_detect.write_log(*args, **kvargs)

    def find_neurons_projections_in_hyper_voxels(self):

        """ For each hyper voxel, list which neurons project to that hypervoxel"""

        if self.role != "master":
            # Master node sets this up
            return

        try:
            # In case the user has been naughty and added extra projections in the code.
            self.sync_projection_info()

            ss = np.random.SeedSequence(self.snudda_detect.random_seed + 202222)
            all_seeds = ss.generate_state(len(self.snudda_detect.neurons))
            all_neuron_id = sorted([n["neuron_id"] for n in self.snudda_detect.neurons])

            seed_lookup = dict()
            for s, n in zip(all_seeds, all_neuron_id):
                seed_lookup[n] = s

            proj_list = []

            self.write_log("!! about to loop projections")

            for neuron_type in self.projections:

                neuron_id_list = self.get_neurons_of_type(neuron_type)
                for neuron_id in neuron_id_list:
                    proj_list.append((neuron_id, neuron_type, seed_lookup[neuron_id]))

            self.hyper_voxel_projections = dict()

            if self.d_view is not None:
                self.d_view.scatter("proj_list", proj_list, block=True)

                cmd_str = "neuron_hv_list = spd.find_hyper_voxel_helper_parallel(proj_info=proj_list)"
                self.d_view.execute(cmd_str, block=True)

                neuron_hv_list = self.d_view.gather("neuron_hv_list", block=True)  # List of list of tuples

                for n_hv in neuron_hv_list:
                    neuron_id = n_hv[0]
                    neuron_hv = n_hv[1]

                    for hid in neuron_hv:
                        if hid not in self.hyper_voxel_projections:
                            self.hyper_voxel_projections[hid] = set([neuron_id])
                        else:
                            self.hyper_voxel_projections[hid].update([neuron_id])

            else:
                neuron_hv_list = self.find_hyper_voxel_helper_parallel(proj_list)

                for n_hv in neuron_hv_list:
                    neuron_id = n_hv[0]
                    neuron_hv = n_hv[1]

                    for hid in neuron_hv:
                        if hid not in self.hyper_voxel_projections:
                            self.hyper_voxel_projections[hid] = set([neuron_id])
                        else:
                            self.hyper_voxel_projections[hid].update([neuron_id])

            # Make sure the workers get the updated info
            self.sync_projection_info()

        except:
            # TODO: Remove this logging code
            import traceback
            t_str = traceback.format_exc()
            self.write_log(t_str, is_error=True)
            print(t_str)
            import pdb
            pdb.set_trace()

    def find_hyper_voxel_helper_parallel(self, proj_info):

        neuron_hv = []

        for neuron_id, neuron_type, random_seed in proj_info:
            n_id, hyper_voxels = self.find_hyper_voxel_helper(neuron_id=neuron_id, neuron_type=neuron_type,
                                                              random_seed=random_seed)

            assert n_id == neuron_id
            neuron_hv.append((n_id, hyper_voxels))

        return neuron_hv

    def find_hyper_voxel_helper(self, neuron_id, neuron_type, random_seed):

        """ Returns hyper voxels that neuron_id (of neuron_type) projects to. """

        try:
            rng = np.random.default_rng(random_seed)

            hyper_voxels = set()

            if neuron_type in self.projections:
                for proj in self.projections[neuron_type]:
                    target_pos, target_rotation, axon_dist = proj["target_info"][neuron_id]
                    num_points = proj["num_points"]

                    if isinstance(proj["radius"], (np.ndarray, list)):
                        rx, ry, rz = proj["radius"]
                    else:
                        rx = ry = rz = proj["radius"]

                    # Find better way to calculate intersection between ellipsoid and hyper voxel cubes?
                    pos = self.ellipsoid_coordinates(target_pos, rx, ry, rz, target_rotation, num_points, rng)

                    hv_idx = ((pos - self.snudda_detect.simulation_origo)
                              / (self.snudda_detect.hyper_voxel_size * self.snudda_detect.voxel_size)).astype(int)

                    hv_id = map(lambda x, y, z: self.snudda_detect.hyper_voxel_id_lookup[x, y, z],
                                hv_idx[:, 0], hv_idx[:, 1], hv_idx[:, 2])

                    hyper_voxels.update(hv_id)

        except:
            import traceback
            t_str = traceback.format_exc()
            self.write_log(t_str, is_error=True)

        return neuron_id, hyper_voxels

    @staticmethod
    def ellipsoid_coordinates(target_pos, rx, ry, rz, rotation, num_points, rng):

        """
            Return num points centred around target_pos, within an ellipsoid with axis rx, ry, rz rotated by rotation

            Args:
                target_pos : 3 x 1 numpy array
                rx, ry, rz : float values, axis of ellipsoid (before rotation)
                rotation : 3x3 rotation matrix, or None
                num_points : number of points to place
                rng : numpy random stream

            Returns:
                coordinates in meters
        """

        rnd_values = rng.random((num_points, 3))  # Use one call for all random values
        theta = 2 * np.pi * rnd_values[:, 0]
        phi = np.arccos(2 * rnd_values[:, 1] - 1)
        r_scale = rnd_values[:, 2] ** (1 / 3)

        x_coord = np.multiply(rx * r_scale, np.multiply(np.sin(phi), np.cos(theta)))
        y_coord = np.multiply(ry * r_scale, np.multiply(np.sin(phi), np.sin(theta)))
        z_coord = np.multiply(rz * r_scale, np.cos(phi))

        pos_offset = np.vstack((x_coord, y_coord, z_coord))

        if rotation is not None:
            pos_offset = np.matmul(rotation, pos_offset)

        pos = target_pos + pos_offset.T

        return pos

    def voxelise_projections(self):

        """ Add the projection of each neuron in pre_neuron_list to the axon space in currently active hyper voxel
         """

        if self.snudda_detect.hyper_voxel_id not in self.hyper_voxel_projections:
            # No projections in this hyper voxel, skip it
            return

        rng = self.snudda_detect.hyper_voxel_rng

        pre_neuron_list = np.sort([x for x in self.hyper_voxel_projections[self.snudda_detect.hyper_voxel_id]])

        for neuron_id in pre_neuron_list:
            neuron_type = self.snudda_detect.neurons[neuron_id]["type"]

            for proj in self.projections[neuron_type]:
                target_pos, target_rotation, axon_dist = proj["target_info"][neuron_id]
                num_points = proj["num_points"]
                if isinstance(proj["radius"], (np.ndarray, list)):
                    rx, ry, rz = proj["radius"]
                else:
                    rx = ry = rz = proj["radius"]

                pos = self.ellipsoid_coordinates(target_pos, rx, ry, rz, target_rotation, num_points, rng)

                # Convert to coordinates in the hyper voxel
                voxel_coords = np.round((pos - self.snudda_detect.hyper_voxel_origo)
                                        / self.snudda_detect.voxel_size).astype(int)

                inside_idx = np.where(np.sum(np.logical_and(0 <= voxel_coords,
                                                            voxel_coords < self.snudda_detect.hyper_voxel_size),
                                             axis=1) == 3)[0]

                for x, y, z in voxel_coords[inside_idx, :]:
                    if self.snudda_detect.axon_voxel_ctr[x, y, z] < self.snudda_detect.max_axon:
                        self.snudda_detect.axon_voxels[x, y, z, self.snudda_detect.axon_voxel_ctr[x, y, z]] = neuron_id
                        self.snudda_detect.axon_voxel_ctr[x, y, z] += 1
                        self.snudda_detect.axon_soma_dist[x, y, z] = axon_dist  # This is an underestimation
                    else:
                        self.snudda_detect.voxel_overflow_counter += 1

    def get_neurons_of_type(self, neuron_type):
        neuron_id = np.array([x["neuron_id"] for x in self.snudda_detect.neurons if x["type"] == neuron_type])
        return neuron_id

    def get_neuron_type(self, neuron_id):
        return self.snudda_detect.neurons[neuron_id]["type"]

    def get_neuron_positions(self, neuron_id):
        return self.snudda_detect.neuron_positions[neuron_id, :]

    def delete_projection(self, projection_name, pre_neuron_type):

        projection_list = self.projections[pre_neuron_type]
        new_projection_list = []
        for proj in projection_list:
            if proj["name"] != projection_name:
                new_projection_list.append(proj)

        self.projections[pre_neuron_type] = new_projection_list

    def parse_config(self):

        for region_name, region_data in self.snudda_detect.config["regions"].items():
            for con_name, con_info in region_data["connectivity"].items():
                for con_type, con_config in con_info.items():
                    if "projection_config_file" in con_config:
                        pre_neuron_type = con_name.split(",")[0]

                        if "projection_name" in con_config:
                            projection_name = con_config["projection_name"]
                        else:
                            projection_name = f"{con_name},{con_type}"

                        self.add_projection(projection_name=projection_name, pre_neuron_type=pre_neuron_type,
                                            projection_file=snudda_parse_path(con_config["projection_config_file"],
                                                                              snudda_data=self.snudda_detect.snudda_data))

    def add_projection(self, projection_name, pre_neuron_type, projection_file):

        """

            Args:
                projection_name (str) : Name of projection
                pre_neuron_type (str) : Neuron type of presynaptic neuron
                projection_file (str) : Path to projection file, used by griddata to find target region coordinates

            Projection mapping and rotations are specified in the projection_file.

        """

        with open(projection_file, "r") as f:
            projection_data = json.load(f, object_pairs_hook=OrderedDict)

        if projection_name and projection_name in projection_data:
            proj_file_info = projection_data[projection_name]
        else:
            proj_file_info = projection_data

        if "axon_morphology" in proj_file_info:
            self.write_log("Axon morphology projections not handled by projection_detection.py")
            self.write_log("UPDATING THE CODE, MAKE SURE detect.py DOES INCLUDE IT IN NEURON MORPHOLOGIES FOR NORMAL DETECTION")
            self.write_log("This code might become obsolete...")
            return

        assert "source" in proj_file_info, f"'source' must exist in {projection_file}"
        assert "destination" in proj_file_info, f"'destination' must exist in {projection_file}"
        assert "radius" in proj_file_info, f"'radius' must exist in {projection_file}"
        assert "num_points" in proj_file_info, f"'num_points' must exist in {projection_file}"

        proj_info = dict()
        proj_info["name"] = projection_name
        proj_info["source"] = np.array(proj_file_info["source"]) * 1e-6
        proj_info["dest"] = np.array(proj_file_info["destination"]) * 1e-6

        proj_info["radius"] = proj_file_info["radius"]  # Either single radius, or (rx, ry, rz)
        proj_info["num_points"] = proj_file_info["num_points"]  # Number of axon points placed

        if "rotation" in projection_data[projection_name]:
            proj_info["rotation"] = np.array(proj_file_info["rotation"])  # n x 9 matrix

        # Pre-compute target centres for each neuron of pre_neuron_type
        neuron_id = self.get_neurons_of_type(pre_neuron_type)
        pre_positions = self.get_neuron_positions(neuron_id)
        target_centres = griddata(points=proj_info["source"], values=proj_info["dest"],
                                  xi=pre_positions, method="linear")

        axon_dist = np.linalg.norm(target_centres - pre_positions, axis=1)

        if "rotation" in proj_info:
            target_rotation = griddata(points=proj_info["dest"], values=proj_info["rotation"],
                                       xi=target_centres, method="linear")
        else:
            target_rotation = [None for x in neuron_id]

        proj_info["target_info"] = dict()
        for nid, pos, rot, ad in zip(neuron_id, target_centres, target_rotation, axon_dist):
            if rot is not None:
                rot_mat = rot.reshape((3, 3))
                assert np.abs(np.linalg.det(rot_mat) - 1) < 1e-6, \
                    f"Invalid rotation matrix for {projection_name}: {rot_mat}"
                proj_info["target_info"][nid] = pos, rot_mat, ad
            else:
                proj_info["target_info"][nid] = pos, None, ad

        if pre_neuron_type not in self.projections:
            self.projections[pre_neuron_type] = [proj_info]
        else:
            self.projections[pre_neuron_type].append(proj_info)



import json
import os

import h5py
import numpy as np
from scipy.spatial import cKDTree

from argparse import ArgumentParser, RawTextHelpFormatter

from snudda.neurons.neuron_prototype import NeuronPrototype
from snudda.utils.load import SnuddaLoad
from snudda.utils.snudda_path import snudda_simplify_path, snudda_parse_path


class SwapToDegeneratedMorphologies:

    def __init__(self, original_network_file, new_network_file,
                 original_snudda_data_dir, new_snudda_data_dir,
                 original_input_file=None, new_input_file=None,
                 filter_axon=False, forced_param_key=None):

        """ This code replaces the neuron morphologies in the original network with user provided degenerated copies
            of the neurons. The synapses that are on removed dendritic will also be removed.
            The section ID and section X is also updated to match the new degenerated morphologies.

            Args:
                original_network_file (str) : Path to input network-synapses.hdf5
                new_network_file (str) : Path to output network-synapses.hdf5
                original_snudda_data_dir (str) : Path to SNUDDA_DATA for original network
                new_snudda_data_dir (str) : Path to SNUDDA_DATA for new network
            """

        self.original_network_file = original_network_file
        self.new_network_file = new_network_file
        self.new_hdf5 = None

        self.original_snudda_data_dir = original_snudda_data_dir
        self.new_snudda_data_dir = new_snudda_data_dir

        assert self.original_snudda_data_dir != self.new_snudda_data_dir, \
            f"SNUDDA_DATA_DIR should be different for WT and degenerated"

        self.original_input_file = original_input_file
        self.new_input_file = new_input_file
        self.forced_param_key = forced_param_key

        self.original_network_loader = SnuddaLoad(self.original_network_file, load_synapses=False)
        self.old_hdf5 = self.original_network_loader.hdf5_file
        self.old_data = self.original_network_loader.data

        self.morphology_map = dict()
        self.section_lookup = dict()
        self.neuron_cache_key = dict()
        self.neuron_cache_id = dict()
        self.key_lookup = dict()

        self.create_section_lookup()
        self.kd_tree_cache = dict()

        self.filter_axon = filter_axon
        self.old_simulation_origo = self.old_hdf5["meta/simulation_origo"][()]
        self.old_voxel_size = self.old_hdf5["meta/voxel_size"][()]

        self.has_axon_density = np.zeros((self.original_network_loader.data["num_neurons"],), dtype=bool)
        for neuron_id, _ in enumerate(self.original_network_loader.data["neurons"]):
            if "axon_density" in self.original_network_loader.data["neurons"][neuron_id] \
                    and self.original_network_loader.data["neurons"][neuron_id]["axon_density"] is not None:
                self.has_axon_density[neuron_id] = True

    def close(self):

        if self.new_hdf5:
            self.new_hdf5.close()

        if self.old_hdf5:
            self.old_hdf5.close()

    def write_new_network_file(self):

        if not os.path.isdir(os.path.dirname(self.new_network_file)):
            print(f"Creating directory {os.path.dirname(self.new_network_file)}")
            os.mkdir(os.path.dirname(self.new_network_file))

        print(f"Writing new network to {self.new_network_file}")
        self.new_hdf5 = h5py.File(self.new_network_file, "w")
        self.old_hdf5.copy(source=self.old_hdf5["meta"], dest=self.new_hdf5)

        if len(self.new_snudda_data_dir) > len(self.original_snudda_data_dir):
            del self.new_hdf5["meta/snudda_data"]
            self.new_hdf5["meta"].create_dataset("snudda_data", data=self.new_snudda_data_dir)
        else:
            self.new_hdf5["meta/snudda_data"][()] = self.new_snudda_data_dir
        network_group = self.new_hdf5.create_group("network")
        self.old_hdf5.copy(source=self.old_hdf5["network/neurons"], dest=self.new_hdf5["network"])

        # Update parameter keys and morphology keys
        for idx, neuron_id in enumerate(self.new_hdf5["network/neurons/neuron_id"]):
            assert idx == neuron_id, "There should be no gaps in numbering."
            param_key, morph_key, neuron_path, param_id, morph_id = self.find_morpology(neuron_id)

            if self.forced_param_key:
                param_key=self.forced_param_key
            self.new_hdf5[f"network/neurons/parameter_key"][idx] = param_key
            self.new_hdf5[f"network/neurons/morphology_key"][idx] = morph_key
            # self.new_hdf5[f"network/neurons/parameter_id"][idx] = param_id
            # self.new_hdf5[f"network/neurons/morphology_id"][idx] = morph_id

        self.filter_synapses(filter_axon=self.filter_axon)
        self.filter_gap_junctions()

    def synapse_iterator(self, data_type=None, load_synapses=True, synapses=None):

        """ Each iteration will return the synapses between one pair of neurons.

            Normally loads the data from the hdf5 file, but if synapses is given,
            then it will use the provided matrix instead. The synapses argument is used by
            post_degeneration_pruning in swap_to_degenerated_morphology.

        """

        start_idx = 0
        next_idx = 0

        if synapses is None:
            data_loc = "network/synapses"
            if data_type is not None and data_type == "gap_junctions":
                data_loc = "network/gap_junctions"
                num_synapses = self.old_hdf5["network/num_gap_junctions"][()]
            else:
                num_synapses = self.old_hdf5["network/num_synapses"][()]

            # Load synapses into memory
            if load_synapses:
                print("Loading synapses into memory.")
                synapses = self.old_hdf5[data_loc][()].copy()
            else:
                print("Reading synapses from disk (slower)")
                synapses = self.old_hdf5[data_loc]

            assert num_synapses == synapses.shape[0]

        else:
            assert data_type is None, (f"If synapses is given, the data will not be loaded from the hdf5 file, "
                                       f"leave data_type as None")
            num_synapses = synapses.shape[0]

        if num_synapses > 0:
            last_pre = synapses[0, 0]
            last_post = synapses[0, 1]

        while next_idx < num_synapses:
            while next_idx < num_synapses \
                    and synapses[next_idx, 0] == last_pre \
                    and synapses[next_idx, 1] == last_post:
                next_idx += 1
                if next_idx % 1000000 == 0:
                    print(f"{next_idx} / {num_synapses}")

            if next_idx < num_synapses:
                last_pre = synapses[next_idx, 0]
                last_post = synapses[next_idx, 1]

            yield synapses[start_idx:next_idx, :]
            start_idx = next_idx

        print(f"{next_idx} / {num_synapses}")

    def gap_junction_iterator(self, load_gap_junctions=True):

        """ Each iteration will return the gap junctions between one pair of neurons. """

        return self.synapse_iterator(data_type="gap_junctions", load_synapses=load_gap_junctions)

    def filter_synapses(self, filter_axon=False):
        # First version, will keep all synapses in memory to write a more efficient file
        new_synapses = np.zeros(self.old_hdf5["network/synapses"].shape,
                                dtype=self.old_hdf5["network/synapses"].dtype)
        syn_ctr = 0

        for synapses in self.synapse_iterator():

            new_syn = self.filter_synapses_helper(synapses, filter_axon=filter_axon)
            new_synapses[syn_ctr:syn_ctr + new_syn.shape[0]] = new_syn
            syn_ctr += new_syn.shape[0]

        self.new_hdf5["network"].create_dataset("synapses", data=new_synapses[:syn_ctr, :], compression="lzf")

        num_synapses = np.zeros((1,), dtype=np.uint64)
        self.new_hdf5["network"].create_dataset("num_synapses", data=syn_ctr, dtype=np.uint64)

        print(f"Keeping {self.new_hdf5['network/num_synapses'][()]} "
              f"out of {self.old_hdf5['network/num_synapses'][()]} synapses "
              f"({self.new_hdf5['network/num_synapses'][()] / self.old_hdf5['network/num_synapses'][()]*100:.3f} %)")

    def filter_gap_junctions(self):
        # First version, will keep all synapses in memory to write a more efficient file
        new_gap_junctions = np.zeros(self.old_hdf5["network/gap_junctions"].shape,
                                     dtype=self.old_hdf5["network/gap_junctions"].dtype)
        gj_ctr = 0

        for gj in self.gap_junction_iterator():
            new_gj = self.filter_gap_junctions_helper(gj)
            new_gap_junctions[gj_ctr:gj_ctr + new_gj.shape[0]] = new_gj
            gj_ctr += new_gj.shape[0]

        self.new_hdf5["network"].create_dataset("gap_junctions", data=new_gap_junctions[:gj_ctr, :], compression="lzf")

        num_gap_junctions = np.zeros((1,), dtype=np.uint64)
        self.new_hdf5["network"].create_dataset("num_gap_junctions", data=gj_ctr, dtype=np.uint64)

        try:
            print(f"Keeping {self.new_hdf5['network/num_gap_junctions'][()]} "
                  f"out of {self.old_hdf5['network/num_gap_junctions'][()]} gap junctions "
                  f"({self.new_hdf5['network/num_gap_junctions'][()] / max(1, self.old_hdf5['network/num_gap_junctions'][()])*100:.3f} %)")
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

    def get_morphology(self, neuron_id=None, hdf5=None, neuron_path=None, snudda_data=None,
                       parameter_key=None, morphology_key=None,
                       neuron_cache_id=None, neuron_cache_key=None):

        """ If neuron_id is given, that neuron will be loaded."""

        if neuron_cache_id is None:
            neuron_cache_id = self.neuron_cache_id

        if neuron_cache_key is None:
            neuron_cache_key = self.neuron_cache_key

        if neuron_id is not None and neuron_id in neuron_cache_id:
            return neuron_cache_id[neuron_id]

        if neuron_path is not None:
            neuron_path = snudda_parse_path(neuron_path, snudda_data)

        if (neuron_path is not None and parameter_key is not None and morphology_key is not None
                and (neuron_path, parameter_key, morphology_key) in self.neuron_cache_key):
            return self.neuron_cache_key[(neuron_path, parameter_key, morphology_key)]

        if neuron_id is not None and hdf5 is not None:
            neuron_path = snudda_parse_path(SnuddaLoad.to_str(hdf5["network/neurons/neuron_path"][neuron_id]),
                                            snudda_data=snudda_data)
            parameter_key = SnuddaLoad.to_str(hdf5["network/neurons/parameter_key"][neuron_id])
            morphology_key = SnuddaLoad.to_str(hdf5["network/neurons/morphology_key"][neuron_id])
            neuron_name = SnuddaLoad.to_str(hdf5["network/neurons/name"][neuron_id])
        else:
            neuron_name = None

        # assert neuron_path is not None and parameter_key is not None and morphology_key is not None, \
        #     "Either provide neuron_id, hdf5 or the three neuron_path, parameter_key and morphology_key"

        if neuron_id is not None and hdf5:
            pos = hdf5["network/neurons/position"][neuron_id, :]
            rot = hdf5["network/neurons/rotation"][neuron_id, :].reshape(3, 3)
        else:
            pos = None
            rot = None

        # TODO: Check if we need to pass snudda_data here!!!
        neuron_prototype = NeuronPrototype(neuron_path=neuron_path, neuron_name=neuron_name, snudda_data=snudda_data)
        neuron = neuron_prototype.clone(parameter_key=parameter_key, morphology_key=morphology_key,
                                        position=pos, rotation=rot)

        if neuron_id is not None:
            neuron_cache_id[neuron_id] = neuron
        else:
            neuron_cache_key[(neuron_path, parameter_key, morphology_key)] = neuron

        return neuron

    def filter_synapses_helper(self, synapses, filter_axon=False):

        """ This currently only takes DENDRITE degeneration into account.
            AXON degeneration is not taken into account for synapse removal.
        """

        pre_id = synapses[0, 0]
        post_id = synapses[0, 1]

        assert (synapses[:, 0] == pre_id).all()
        assert (synapses[:, 1] == post_id).all()

        old_sec_id = synapses[:, 9]
        old_sec_x = synapses[:, 10]

        keep_idx, new_sec_id, new_sec_x \
            = self.remap_sections_helper(neuron_id=post_id, old_sec_id=old_sec_id, old_sec_x=old_sec_x/1000.0)

        edited_synapses = synapses.copy()

        edited_synapses[:, 9] = new_sec_id
        edited_synapses[:, 10] = new_sec_x * 1000

        if filter_axon:
            filtered_synapses = self.filter_axonal_synapses_helper(edited_synapses[keep_idx, :])
        else:
            filtered_synapses = edited_synapses[keep_idx, :]

        return filtered_synapses

    def filter_axonal_synapses_helper(self, synapses, max_dist=5.41e-6):

        """ Filter the synapses that have the axon degeneration, presynaptic neurons without axons are ignored. """

        pre_id = synapses[:, 0]
        synapse_voxels = synapses[:, 2:5]

        synapse_coordinates = self.old_simulation_origo + self.old_voxel_size*synapse_voxels

        keep_idx = np.ones((synapses.shape[0],), dtype=bool)

        loaded_pid = None
        axon_tree = None

        # Loop through all the synapses, if they have an axon check that there is an axonal point close to synapse
        for idx, (pid, coord) in enumerate(zip(pre_id, synapse_coordinates)):

            if pid != loaded_pid:
                morph = self.get_morphology(neuron_id=pid, hdf5=self.new_hdf5, snudda_data=self.new_snudda_data_dir)
                loaded_pid = pid

                # Has axon?
                if 2 in morph.morphology_data["neuron"].sections and len(morph.morphology_data["neuron"].sections[2]) > 0:
                    axon_tree = self.get_kd_tree(morph, "axon")

                    if self.has_axon_density[pid]:
                        axon_tree = None
                        # print(f"Warning: Axon and axonal density specified for neuron {pid}")
                        # print(f"Ignoring morphology --- for now, will change behaviour in the future.")
                else:
                    axon_tree = None

            if axon_tree is None:
                # No pre-synaptic axon exists for this neuron (no info, so keep synapse)
                continue

            closest_dist, closest_point = axon_tree.query(coord)

            # Here we assume one query per point, so closest_dist should be a scalar
            if closest_dist > max_dist:
                keep_idx[idx] = 0

        return synapses[keep_idx, :].copy()

    def filter_gap_junctions_helper(self, gap_junctions):

        # Gap junctions are bidirectional, but call them pre and post out of habit
        pre_id = gap_junctions[0, 0]
        post_id = gap_junctions[0, 1]

        assert (gap_junctions[:, 0] == pre_id).all()
        assert (gap_junctions[:, 1] == post_id).all()

        old_pre_sec_id = gap_junctions[:, 2]
        old_pre_sec_x = gap_junctions[:, 4]
        old_post_sec_id = gap_junctions[:, 3]
        old_post_sec_x = gap_junctions[:, 5]

        keep_idx_pre, new_pre_sec_id, new_pre_sec_x \
            = self.remap_sections_helper(neuron_id=pre_id, old_sec_id=old_pre_sec_id, old_sec_x=old_pre_sec_x/1000)

        keep_idx_post, new_post_sec_id, new_post_sec_x \
            = self.remap_sections_helper(neuron_id=post_id, old_sec_id=old_post_sec_id, old_sec_x=old_post_sec_x/1000)

        keep_idx = np.intersect1d(keep_idx_pre, keep_idx_post)

        edited_gap_junctions = gap_junctions.copy()
        edited_gap_junctions[:, 2] = new_pre_sec_id
        edited_gap_junctions[:, 3] = new_post_sec_id
        edited_gap_junctions[:, 4] = new_pre_sec_x * 1000
        edited_gap_junctions[:, 5] = new_post_sec_x * 1000

        return edited_gap_junctions[keep_idx, :]

    def find_old_morphology(self, neuron_id):

        if neuron_id not in self.key_lookup:
            orig_morph_key = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["morphology_key"])
            orig_param_key = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["parameter_key"])
            orig_neuron_path = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["neuron_path"])

            self.key_lookup[neuron_id] = orig_param_key, orig_morph_key, orig_neuron_path
        else:
            orig_param_key, orig_morph_key, orig_neuron_path = self.key_lookup[neuron_id]

        return orig_param_key, orig_morph_key, orig_neuron_path

    def find_morpology(self, neuron_id):

        """ Given neuron_id it looks up what was old morphology, then looks for a new morphology with the same name
            and finds the corresponding morphology_key. It also tries to pick a parameter_key """

        orig_morph_key = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["morphology_key"])
        orig_param_key = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["parameter_key"])

        # We assume the new morphology is in the same relative path, but using a different SNUDDA_DATA
        orig_simple_path = SnuddaLoad.to_str(self.old_data["neurons"][neuron_id]["neuron_path"])
        orig_neuron_path = snudda_parse_path(orig_simple_path, os.path.realpath(self.original_snudda_data_dir))
        # new_neuron_path = snudda_parse_path(orig_simple_path, os.path.realpath(self.new_snudda_data_dir))
        new_neuron_path = orig_neuron_path.replace(os.path.realpath(self.original_snudda_data_dir),
                                                   os.path.realpath(self.new_snudda_data_dir))

        if orig_morph_key is None or orig_morph_key == '':
            # Only a single morphology

            original_morphology_id = 0  # self.old_data["neurons"][neuron_id]["morphology_id"]
            original_parameter_id = 0  # self.old_data["neurons"][neuron_id]["parameter_id"]

            return '', '', new_neuron_path, original_parameter_id, original_morphology_id

        # Here we assume there is a meta.json file
        with open(os.path.join(orig_neuron_path, "meta.json"), "r") as f:
            orig_meta_info = json.load(f)

        with open(os.path.join(new_neuron_path, "meta.json"), "r") as f:
            new_meta_info = json.load(f)

        # Find the parameter_key and morphology_key corresponding to the morphology used. Note that there might be
        # multiple parameter_key:s valid for a morphology, so then we randomly pick one of those

        orig_morph_name = orig_meta_info[orig_param_key][orig_morph_key]["morphology"]

        possible_keys = []

        for param_key, param_data in new_meta_info.items():
            for morph_key, morph_data in param_data.items():
                morph_name = morph_data["morphology"]
                if orig_morph_name == morph_name:
                    possible_keys.append((param_key, morph_key))

        assert len(possible_keys) > 0, \
            f"No morphology matching for {orig_morph_name}, unable to pick parameter_key and morphology_key"

        # Pick one of the key pairs
        idx = np.random.randint(low=0, high=len(possible_keys))
        new_param_key, new_morph_key = possible_keys[idx]

        # We also need parameter_id and morphology_id
        parameter_id = np.where([x == new_param_key for x in new_meta_info.keys()])[0][0]
        morphology_id = np.where([x == new_morph_key for x in new_meta_info[new_param_key].keys()])[0][0]

        return new_param_key, new_morph_key, new_neuron_path, parameter_id, morphology_id

    def get_sec_location(self, coords, neuron_path, snudda_data,
                         parameter_key, morphology_key, max_dist=5.41e-6):

        raise DeprecationWarning("This function is no longer used. It is based on old NeuronMorphology -- REMOVE?")
        assert False, "Do not run this!"

        morph = self.get_morphology(neuron_path=neuron_path,
                                    parameter_key=parameter_key,
                                    morphology_key=morphology_key,
                                    snudda_data=snudda_data)

        dend = self.get_kd_tree(morph, "dend")
        sec_id = np.zeros((dend.shape[0],), dtype=int)
        sec_x = np.zeros((dend.shape[0],))

        coord_to_sec_id_x = dict()
        for link, sec_id, sec_x in zip(morph.dend_links, morph.dend_sec_id, morph.dend_sec_x):
            coord = morph.dend[link[1], :]

            coord_to_sec_id_x[coord] = (sec_id, sec_x[1])

        for idx, coord in enumerate(coords):
            closest_dist, closest_point = dend.query(coord)

            if closest_dist <= max_dist:
                syn_sec_id, syn_sec_x = coord_to_sec_id_x[morph.dend[closest_point, :3]]
                sec_id[idx] = syn_sec_id
                sec_x[idx] = syn_sec_x
            else:
                sec_id[idx] = np.nan
                sec_x[idx] = np.nan

        return sec_id, sec_x

    def write_new_input_file(self, remap_removed_input=False, remapped_fraction=1.0):

        if self.original_input_file is None:
            print("No input file supplied, no new input file created")
            return

        if not os.path.isdir(os.path.dirname(self.new_input_file)):
            os.mkdir(os.path.dirname(self.new_input_file))

        print(f"Writing new input data to {self.new_input_file}")

        old_input = h5py.File(self.original_input_file, "r")
        new_input = h5py.File(self.new_input_file, "w")
        old_input.copy(source=old_input["config"], dest=new_input)

        new_input.create_group("input")

        for neuron in old_input["input"].keys():
            neuron_group = new_input["input"].create_group(neuron)

            old_n = 0
            new_n = 0
            remap_n = 0

            for input_type in old_input["input"][neuron].keys():
                # Note: This code assumes these are real neuron and not just virtual neurons.
                #       it does not handle the virtual neuron case here.
                old_input_data = old_input["input"][neuron][input_type]

                keep_idx, new_sec_id, new_sec_x \
                    = self.remap_sections_helper(neuron_id=int(neuron),
                                                 old_sec_id=old_input_data.attrs["section_id"],
                                                 old_sec_x=old_input_data.attrs["section_x"])

                if len(keep_idx) == 0 and not (remap_removed_input and remapped_fraction > 0):
                    continue

                input_group = neuron_group.create_group(input_type)

                if remap_removed_input:

                    # We need to find new positions for input marked as removed
                    morph = self.get_morphology(neuron_id=int(neuron), hdf5=self.new_hdf5,
                                                snudda_data=self.new_snudda_data_dir)

                    n_remap = len(old_input_data.attrs["section_id"]) - len(keep_idx)
                    idx_remap = sorted(list(set(np.arange(0, len(old_input_data.attrs["section_id"]))) - set(keep_idx)))

                    if remapped_fraction < 1.0:
                        n_remap = int(np.round(len(idx_remap) * remapped_fraction))
                        idx_remap = sorted(list(np.random.permutation(idx_remap)[:n_remap]))

                    try:
                        synapse_density = SnuddaLoad.to_str(old_input_data.attrs["synapse_density"])
                    except:
                        import traceback
                        print(traceback.format_exc())
                        import pdb
                        pdb.set_trace()

                    xyz, sec_id, sec_x, dist_to_soma = morph.dendrite_input_locations(synapse_density_str=synapse_density,
                                                                                      num_locations=n_remap,
                                                                                      rng=self.rng,
                                                                                      cluster_size=1,
                                                                                      cluster_spread=None)

                    old_n += old_input_data['spikes'].shape[0]
                    new_n += len(keep_idx)
                    remap_n += len(idx_remap)

                    keep_idx2 = sorted(list(set(keep_idx).union(set(idx_remap))))

                    # Same spikes as before
                    spike_set = input_group.create_dataset("spikes", data=old_input_data["spikes"][keep_idx2, :],
                                                           compression="gzip", dtype=np.float32)
                    spike_set.attrs["num_spikes"] = old_input_data["spikes"].attrs["num_spikes"][keep_idx2].astype(np.int32)

                    # New locations for the remapped synapses
                    new_sec_id[idx_remap] = sec_id
                    input_group.attrs["section_id"] = new_sec_id[keep_idx2].astype(np.int16)

                    new_sec_x[idx_remap] = sec_x
                    input_group.attrs["section_x"] = new_sec_x[keep_idx2].astype(np.float16)

                    input_group.attrs["parameter_id"] = old_input_data.attrs["parameter_id"][keep_idx2].astype(np.int)

                    updated_dist = old_input_data.attrs["distance_to_soma"].copy()
                    updated_dist[idx_remap] = dist_to_soma
                    input_group.attrs["distance_to_soma"] = updated_dist[keep_idx2].astype(np.float16)

                    for data_name in ["freq", "correlation", "jitter", "synapse_density", "start", "end", "conductance",
                                      "population_unit_id", "population_unit_spikes", "generator",
                                      "mod_file", "parameter_file", "parameter_list"]:

                        if data_name in old_input_data.attrs:
                            input_group.attrs[data_name] = old_input_data.attrs[data_name]
                        elif data_name in old_input_data["spikes"].attrs:
                            input_group["spikes"].attrs[data_name] = old_input_data["spikes"].attrs[data_name]
                        elif data_name in old_input_data:
                            old_input_data.copy(source=old_input_data[data_name], dest=input_group)

                else:
                    input_group.create_dataset("spikes", data=old_input_data["spikes"][keep_idx, :],
                                               compression="gzip", dtype=np.float32)
                    input_group["spikes"].attrs["num_spikes"] = old_input_data["spikes"].attrs["num_spikes"][keep_idx]
                    input_group.attrs["section_id"] = new_sec_id[keep_idx]
                    input_group.attrs["section_x"] = new_sec_x[keep_idx]
                    input_group.attrs["parameter_id"] = old_input_data.attrs["parameter_id"][keep_idx]
                    input_group.attrs["distance_to_soma"] = old_input_data.attrs["distance_to_soma"][keep_idx]

                    for data_name in ["freq", "correlation", "jitter", "synapse_density", "start", "end", "conductance",
                                      "population_unit_id", "population_unit_spikes", "generator",
                                      "mod_file", "parameter_file", "parameter_list"]:

                        if data_name in old_input_data.attrs:
                            input_group.attrs[data_name] = old_input_data.attrs[data_name]
                        elif data_name in old_input_data["spikes"].attrs:
                            input_group["spikes"].attrs[data_name] = old_input_data["spikes"].attrs[data_name]
                        elif data_name in old_input_data:
                            old_input_data.copy(source=old_input_data[data_name], dest=input_group)

                    old_n += old_input_data['spikes'].shape[0]
                    new_n += len(keep_idx)

            print(f"Processed input to {self.old_data['neurons'][int(neuron)]['name']} ({neuron}), "
                  f"keeping {new_n} out of {old_n} inputs (plus remapping {remap_n} inputs)"
                  f"({new_n / max(old_n,1) * 100 :.2f} %) ({remap_n / max(old_n,1)*100 :.2f} % remapped)")

        self.close()

    def get_kd_tree(self, neuron, tree_type, kd_tree_cache=None):

        if kd_tree_cache is None:
            kd_tree_cache = self.kd_tree_cache

        morph_type_lookup = {"axon": 2, "dend": 3}

        if (neuron, tree_type) not in kd_tree_cache:
            morph_type = morph_type_lookup[tree_type]

            idx = np.where(neuron.morphology_data["neuron"].section_data[:, 2] == morph_type)[0]
            coords = neuron.morphology_data["neuron"].geometry[idx, :3]

            if coords.size > 0:
                kd_tree_cache[(neuron, tree_type)] = cKDTree(coords)
            else:
                kd_tree_cache[(neuron, tree_type)] = None

        return kd_tree_cache[(neuron, tree_type)]

    def create_section_lookup(self):

        for neuron_id in self.old_hdf5["network/neurons/neuron_id"]:
            self.create_section_lookup_helper(neuron_id)

    # Note this does not take AXON degeneration into account, need to do that separately
    def create_section_lookup_helper(self, neuron_id):

        old_param_key, old_morph_key, old_path = self.find_old_morphology(neuron_id=neuron_id)

        new_param_key, new_morph_key, new_path, _, _ = self.find_morpology(neuron_id=neuron_id)

        if (old_param_key, old_morph_key, old_path) in self.section_lookup:
            # We already have the morphology in the section_lookup
            return

        old_morph = self.get_morphology(parameter_key=old_param_key, morphology_key=old_morph_key,
                                        neuron_path=old_path, snudda_data=self.original_snudda_data_dir)
        new_morph = self.get_morphology(parameter_key=new_param_key, morphology_key=new_morph_key,
                                        neuron_path=new_path, snudda_data=self.new_snudda_data_dir)
        coord_to_sec_id_x = dict()
        old_dend_idx = np.where(old_morph.morphology_data["neuron"].section_data[:, 2] == 3)[0]
        for idx in old_dend_idx:
            coord = (old_morph.morphology_data["neuron"].geometry[idx, :3] * 1e9).astype(int)
            assert (coord[0], coord[1], coord[2]) not in coord_to_sec_id_x, \
                f"Coordinates {(coord[0], coord[1], coord[2])} already exists in {coord_to_sec_id_x}"

            old_sec_id = old_morph.morphology_data["neuron"].section_data[idx, 0]
            old_sec_x = old_morph.morphology_data["neuron"].section_data[idx, 1] / 1e3
            coord_to_sec_id_x[coord[0], coord[1], coord[2]] = (old_sec_id, old_sec_x)

        old_to_new_sec_id = dict()
        old_sec_x_list = dict()
        new_sec_x_list = dict()
        # We just need to find the maximal old sec_x still present,
        # that value will map to sec_x 1.0 in new (stored as int sec_x*1000)

        new_dend_idx = np.where(new_morph.morphology_data["neuron"].section_data[:, 2] == 3)[0]
        for idx in new_dend_idx:
            coord = (new_morph.morphology_data["neuron"].geometry[idx, :3] * 1e9).astype(int)
            new_sec_id = new_morph.morphology_data["neuron"].section_data[idx, 0]
            new_sec_x = new_morph.morphology_data["neuron"].section_data[idx, 1] / 1e3

            try:
                old_sec_id, old_sec_x = coord_to_sec_id_x[coord[0], coord[1], coord[2]]
            except:
                import traceback
                print(traceback.format_exc())
                print(f"Coordinate point in new file missing in old file.\n"
                      f"Old morphology: {old_morph.swc_filename}\n"
                      f"New morphology: {new_morph.swc_filename}")
                import pdb
                pdb.set_trace()

            # TODO: What happens if a branch looses one of its side branches, will then those two sections
            #       be merged into one? This code will then complain.

            if old_sec_id in old_to_new_sec_id:
                assert old_to_new_sec_id[old_sec_id] == new_sec_id, \
                    (f"Old sec_id {old_sec_id} maps to multiple " 
                     f"new sec_id {new_sec_id} and {old_to_new_sec_id[old_sec_id]}")
            else:
               old_to_new_sec_id[old_sec_id] = new_sec_id

            if old_sec_id not in old_sec_x_list:
                old_sec_x_list[old_sec_id] = [old_sec_x]
                new_sec_x_list[old_sec_id] = [new_sec_x]
            else:
                old_sec_x_list[old_sec_id].append(old_sec_x)
                new_sec_x_list[old_sec_id].append(new_sec_x)

        new_to_old_sec_id = dict()
        for old_sec_id, new_sec_id in old_to_new_sec_id.items():
            if new_sec_id not in new_to_old_sec_id:
                new_to_old_sec_id[new_sec_id] = [old_sec_id]
            else:
                new_to_old_sec_id[new_sec_id].append(old_sec_id)

        for old_sec_id, new_sec_id in old_to_new_sec_id.items():
            # Is smallest section_x not zero?
            if old_sec_x_list[old_sec_id][0] > 0:
                # Are there multiple sections mapping to this new section_id?
                if len(new_to_old_sec_id[new_sec_id]) == 1 or old_sec_id == min(new_to_old_sec_id[new_sec_id]):
                    # Either just one section mapping to the new section_id
                    # or this is the smallest old section_id mapping to the new section_id
                    old_sec_x_list[old_sec_id].insert(0, 0)
                    new_sec_x_list[old_sec_id].insert(0, 0)

        # Soma ID now -1, updated mapping
        neuron_section_lookup = {-1: (-1, np.array([0, 1]), np.array([0, 1]))}  # Add SOMA mapping. SecX 0-1 --> ID 0-1

        for old_sec_id in old_to_new_sec_id.keys():

            assert (np.diff(old_sec_x_list[old_sec_id]) > 0).all()
            assert (np.diff(new_sec_x_list[old_sec_id]) > 0).all()

            neuron_section_lookup[old_sec_id] = (old_to_new_sec_id[old_sec_id],
                                                 np.array(old_sec_x_list[old_sec_id]),
                                                 np.array(new_sec_x_list[old_sec_id]))

        if True:
            assert len(neuron_section_lookup) > 3, (f"Section lookup has few elements. Does morphologies match?"
                                                     f"\nOld = {old_path, old_param_key, old_morph_key}"
                                                     f"\nNew = {new_path, new_param_key, new_morph_key}")

            # Check that all new_coords exist in the old_coords list.

        # import pdb
        # pdb.set_trace()

        self.section_lookup[old_param_key, old_morph_key, old_path] = neuron_section_lookup

    def remap_sections_helper(self, neuron_id, old_sec_id, old_sec_x):

        """
            Remaps sections for neuron_id.

            This does currently NOT take AXON DEGENERATION into account.

            Args:
                neuron_id (int) : Neuron ID.
                old_sec_id (np.array) : Section ID of old section.
                old_sec_x (np.array) : Section X of old section (float 0-1).

            Returns:
                keep_idx (np.array) : Indices to keep.
                new_sec_id (np.array) : Section ID of new section, positions are np.nan if not mapped.
                new_sec_x (np.array) : Section X of new section, positions are np.nan if not mapped.

        """

        keep_idx = np.zeros(old_sec_id.shape, dtype=bool)
        new_sec_id = np.full(old_sec_id.shape, -9999999, dtype=int)
        new_sec_x = np.full(old_sec_id.shape, np.nan)

        old_param_key, old_morph_key, old_path = self.find_old_morphology(neuron_id=neuron_id)

        lookup = self.section_lookup[old_param_key, old_morph_key, old_path]

        for idx, (old_id, old_x) in enumerate(zip(old_sec_id, old_sec_x)):

            if old_id in lookup:
                new_id = lookup[old_id][0]

                # We know range of old x that are still valid, and new x they map to. If outside, returns nan
                try:
                    new_x = np.interp(old_x, lookup[old_id][1], lookup[old_id][2], right=np.nan, left=np.nan)
                except:
                    import traceback
                    print(traceback.format_exc())
                    import pdb
                    pdb.set_trace()

                if not np.isnan(new_x):
                    assert 0 <= new_x <= 1, f"Out of range new_x={new_x} (required 0-1)"

                    keep_idx[idx] = True
                    new_sec_id[idx] = new_id
                    new_sec_x[idx] = new_x

        return np.where(keep_idx)[0], new_sec_id, new_sec_x


def cli():
    parser = ArgumentParser(
        description="Replace WT morphologies with PD morphologies, remove orphaned synapses and input",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("original_network_path", type=str)
    parser.add_argument("new_network_path", type=str)
    parser.add_argument("original_snudda_path", type=str)
    parser.add_argument("new_snudda_path", type=str)
    parser.add_argument("--remap", action="store_true")

    args = parser.parse_args()

    if not os.path.isdir(args.new_network_path):
        os.mkdir(args.new_network_path)

    original_network_file = os.path.join(args.original_network_path, "network-synapses.hdf5")
    new_network_file = os.path.join(args.new_network_path, "network-synapses.hdf5")
    original_input_file = os.path.join(args.original_network_path, "input-spikes.hdf5")
    new_input_file = os.path.join(args.new_network_path, "input-spikes.hdf5")

    swap = SwapToDegeneratedMorphologies(original_network_file=original_network_file,
                                         new_network_file=new_network_file,
                                         original_snudda_data_dir=args.original_snudda_path,
                                         new_snudda_data_dir=args.new_snudda_path,
                                         original_input_file=original_input_file,
                                         new_input_file=new_input_file)
    swap.write_new_network_file()
    swap.write_new_input_file(remap_removed_input=args.remap)

    swap.close()


if __name__ == "__main__":
    cli()



# This script visualises two neurons and the location of the synapses
# connecting them.

print("!!! THIS CODE IS REALLY SLOW !!! -- use makeNeuronCube.py if you only want neurons without synapses")

import bpy
import mathutils
import numpy as np
import h5py
from snudda.utils.snudda_path import snudda_parse_path

# networkFile = "/home/hjorth/HBP/StriatumNetwork/model/SmallTest2/network-connect-voxel-pruned-synapse-file.hdf5"
networkFile = "/home/hjorth/HBP/StriatumNetwork/model/SmallTest2/network-connect-voxel-pruned-synapse-file.hdf5"

outFile = "cubeOfNeurons-synapses-white.blend"

visualiseID = range(0,1800) #[20,1217]

# Load the neuron positions
fi = h5py.File(networkFile,"r")

neuronID = fi["network/neurons/neuronID"].value
keepIdx = np.array([(x in visualiseID) for x in neuronID])
neuronID = neuronID[keepIdx]

pos = fi["network/neurons/position"][keepIdx,:]
rot = fi["network/neurons/rotation"][keepIdx,:]
morph = fi["network/neurons/morphology"][keepIdx]
name = fi["network/neurons/name"][keepIdx]

origo = fi["meta/simulationOrigo"].value
voxelSize = fi["meta/voxelSize"].value


synapses = fi["network/synapses"]


# Remove the start cube
bpy.ops.object.delete() 

bpy.data.scenes['Scene'].render.engine = 'CYCLES'
world = bpy.data.worlds['World']
world.use_nodes = True

whiteBackground = True

# changing these values does affect the render.
bg = world.node_tree.nodes['Background']
if(whiteBackground): # Set to True for white background
  bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
  bg.inputs[1].default_value = 1.0
else:
  bg.inputs[0].default_value[:3] = (0.0, 0.0, 0.0)
  bg.inputs[1].default_value = 0.0

  
matMSD1 = bpy.data.materials.new("PKHG")
#matMSD1.diffuse_color = (1.0,0.2,0.2)
matMSD1.diffuse_color = (77./255,151./255,1.0)

matMSD2 = bpy.data.materials.new("PKHG")
#matMSD2.diffuse_color = (0.2,0.2,1.0)
matMSD2.diffuse_color = (67./255,55./255,181./255)

matFS = bpy.data.materials.new("PKHG")
matFS.diffuse_color = (6./255,31./255,85./255)
matChIN = bpy.data.materials.new("PKHG")
matChIN.diffuse_color = (252./255,102./255,0.0)
matLTS = bpy.data.materials.new("PKHG")
matLTS.diffuse_color = (150./255,63./255,212./255)

matOther = bpy.data.materials.new("PKHG")
matOther.diffuse_color = (0.4,0.4,0.4)

matSynapse = bpy.data.materials.new("PKHG")
if(whiteBackground):
  matSynapse.diffuse_color = (0.8,0.0,0.0)
else:
  matSynapse.diffuse_color = (1.0,1.0,0.9)
  
#matSynapse.use_transparency = True
matSynapse.use_nodes = True
#import pdb
#pdb.set_trace()


if(not whiteBackground):
  emissionStrength = 5.0
  
  # Make synapses glow
  emission = matSynapse.node_tree.nodes.new('ShaderNodeEmission')
  emission.inputs['Strength'].default_value = emissionStrength

  material_output = matSynapse.node_tree.nodes.get('Material Output')
  matSynapse.node_tree.links.new(material_output.inputs[0], emission.outputs[0])


for ps,rt,mo,nm,nID in zip(pos,rot,morph,name,neuronID):

  eRot = mathutils.Matrix(rt.reshape(3,3)).to_euler()
  bpy.ops.import_mesh.swc(filepath=snudda_parse_path(mo))
  obj = bpy.context.selected_objects[0]
  
  obj.rotation_euler = eRot
  print("Setting position: " + str(ps*1e3))
  obj.location = ps*1e3

  nType = nm.decode().split("_")[0]
  if(nType == "dSPN" or nType == "MSD1"):
    print(str(nID) + " dSPN")
    mat = matMSD1
  elif(nType == "iSPN" or nType == "MSD2"):
    print(str(nID) + " iSPN")
    mat = matMSD2
  elif(nType == "FS"):
    print(str(nID) + " FS")
    mat = matFS
  elif(nType == "ChIN"):
    print(str(nID) + " ChIN")
    mat = matChIN
  elif(nType == "LTS"):
    print(str(nID) + " LTS")
    mat = matLTS
  else:
    print(str(nID) + " other")
    mat = matOther

  for ch in obj.children:
    ch.active_material = mat

  obj.select = False


# Draw the synapses

drawSynapses = False

if(drawSynapses):

  nSynapses = 0

  for ob in bpy.context.selected_objects:
    ob.select = False

  for syn in synapses:
    preID = syn[0]
    postID = syn[1]

    if(preID in visualiseID and postID in visualiseID):
      # Draw this neuron (the SWC import scales from micrometers to mm), the
      # positions in the simulation are in meters, need to scale it to mm for
      # blender to have same units.
      x = (origo[0] + voxelSize*syn[2]) * 1e3
      y = (origo[1] + voxelSize*syn[3]) * 1e3
      z = (origo[2] + voxelSize*syn[4]) * 1e3

      #bpy.ops.mesh.primitive_cube_add(location=(x,y,z))
      #bpy.context.scene.objects.active.dimensions = \
        #  (voxelSize, voxelSize, voxelSize)

      bpy.ops.mesh.primitive_uv_sphere_add(location=(x,y,z),size=0.001*4)
      ob = bpy.context.selected_objects[0] 
      ob.active_material = matSynapse
      ob.select = False
      
      nSynapses += 1
    
      print("Added synapse at " + str([x,y,z]))

  print("nSynapses = " + str(nSynapses))
    
#import pdb
#pdb.set_trace()

# Add a light source

lampData = bpy.data.lamps.new(name="Sun", type='SUN')
lampObject = bpy.data.objects.new(name="Sun", object_data=lampData)
bpy.context.scene.objects.link(lampObject)

# Place lamp to a specified location
lampObject.location = (1000.0, 1000.0, 1000.0)

# Reposition camera
cam = bpy.data.objects["Camera"]
#cam.location = (2.98,2.68,4.96)
#cam.rotation_euler = (1.59,0,-0.26)

cam.location = (3.19,3.46,4.96)
cam.rotation_euler = (96.7*np.pi/180,0,-14.3*np.pi/180)


if(True):
  bpy.ops.render.render()
  bpy.data.images['Render Result'].save_render(filepath="cube-of-neurons-with-neurites.png")

bpy.ops.wm.save_as_mainfile(filepath=outFile)



import matplotlib.pyplot as plt
import numpy as np
from snudda.utils import SnuddaLoad


class PlotConnectionMatrix:

    def __init__(self, network_path):

        self.network_path = network_path
        self.snudda_load = SnuddaLoad(network_path)

        print("Creating connection matrix...")
        self.con_mat = self.snudda_load.create_connection_matrix(sparse_matrix=False)

    def plot_sorted_connection_matrix(self, fig_path, neuron_order=None):

        all_neuron_id = []

        if neuron_order is None:
            neuron_order = sorted(list(set(self.snudda_load.get_neuron_types())))

        print(f"neuron_order = {neuron_order}")

        label_edges = [0]
        label_centers = []
        labels = []

        for neuron_type in neuron_order:
            print(f"Plotting {neuron_type}")
            neuron_id = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type)
            all_neuron_id.append(neuron_id)
            label_centers.append(label_edges[-1] + len(neuron_id)/2)
            label_edges.append(len(neuron_id)+label_edges[-1])
            labels.append(neuron_type)

        all_neuron_id = np.concatenate(all_neuron_id)

        sorted_con_mat = self.con_mat[all_neuron_id, :][:, all_neuron_id]

        cbar_kw = {}
        cbarlabel = ""

        neuron_types = self.snudda_load.get_neuron_types(neuron_id=all_neuron_id)

        fig, ax = plt.subplots()
        # cmap: 'Spectral', 'coolwarm', 'jet
        im = ax.imshow(sorted_con_mat, cmap='Spectral')

        cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
        cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

        ax.set_xticks(label_centers, labels=labels)
        ax.set_yticks(label_centers, labels=labels)

        #ax.set_xticks(label_edges, minor=True)
        #ax.set_yticks(label_edges, minor=True)
        
        ax.set_xlabel('Postsynaptic neurons')
        ax.set_ylabel('Presynaptic neurons')


        fig.tight_layout()
        print(f"Saving figure to {fig_path}")
        plt.savefig(fig_path)
        plt.show()


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser("Plot connection matrix")
    parser.add_argument("network_path", help="Path to network folder", type=str)
    parser.add_argument("output_file", help="Figure file", type=str)
    parser.add_argument("--order", help="e.g. dSPN,iSPN,FS (no spaces)", type=str)

    args = parser.parse_args()

    if args.order is not None:
        type_order = args.order.split(",")
    else:
        type_order = None

    print(f"type_order: {type_order}")

    pc = PlotConnectionMatrix(network_path=args.network_path)
    pc.plot_sorted_connection_matrix(neuron_order=type_order, fig_path=args.output_file)




#!/usr/bin/env python3

import numpy as np

from snudda.utils.export_connection_matrix import SnuddaExportConnectionMatrix


class SnuddaExportErodedConnectionMatrix(SnuddaExportConnectionMatrix):

    def __init__(self, in_file, out_file, fraction_kept=1.0, save_sparse=True, permute=False, permute_type="types"):
        super().__init__(in_file=in_file, out_file=out_file, save_sparse=save_sparse, save_on_init=False)

        self.con_mat = self.erode(mat=self.con_mat, fraction_kept=fraction_kept)

        if permute:
            if permute_type == "all":
                self.con_mat = self.permute_all(self.con_mat)
            elif permute_type == "types":
                self.con_mat = self.permute_within_neuron_types(self.con_mat)
            elif permute_type == "noself":
                self.con_mat = self.permute_all_no_self_connections(self.con_mat)
            else:
                raise ValueError(f"Unknown permutation type {permute_type}, please use 'all', 'types', 'noself'")

        self.save()

    @staticmethod
    def erode(mat, fraction_kept=1.0):
        erode_mat = np.random.uniform(size=mat.shape) <= fraction_kept
        mat = np.multiply(erode_mat, mat)

        return mat

    @staticmethod
    def permute_all(mat):
        row_permute = np.random.permutation(mat.shape[0])
        col_permute = np.random.permutation(mat.shape[1])

        return mat[row_permute, :][:, col_permute]

    def permute_all_no_self_connections(self, mat):

        s = mat.shape
        assert s[0] == s[1] and len(s) == 2

        non_diagonal_idx = np.where(np.diag(np.ones((s[0],))).flatten() == 0)
        permuted_idx = np.random.permutation(non_diagonal_idx)

        new_mat = mat.flatten()

        new_mat[non_diagonal_idx] = mat.flatten()[permuted_idx]
        new_mat = new_mat.reshape(mat.shape)

        return new_mat

    def get_id_of_all_types(self):

        idx = []
        for nt in set(self.sl.get_neuron_types()):
            idx.append(self.sl.get_neuron_id_of_type(neuron_type=nt))

        return idx

    def permute_within_neuron_types(self, mat, list_of_type_idx=None):

        """ Args:
                mat : Connection matrix
                list_of_type_idx : [(1,2,3,4), (5,6,7,8), (9,10)] if neurons of 1,2,3,4 are same type, 5,6,7,8 same
                                   and 9, 10 are of same type.
        """

        if list_of_type_idx is None:
            list_of_type_idx = self.get_id_of_all_types()

        all_idx = np.concatenate(list_of_type_idx)
        all_idx_sorted = np.sort(all_idx)
        assert (np.diff(all_idx_sorted) == 1).all()
        assert all_idx_sorted[0] == 0 and len(all_idx_sorted) == mat.shape[0] == mat.shape[1]

        new_mat = np.full(mat.shape, np.nan).flatten()

        for type_ctr_pre, pre_idx in enumerate(list_of_type_idx):
            for type_ctr_post, post_idx in enumerate(list_of_type_idx):

                flat_idx = np.ravel_multi_index(np.ix_(pre_idx, post_idx), mat.shape).flatten()

                if type_ctr_pre == type_ctr_post:
                    new_mat[flat_idx] = self.permute_all_no_self_connections(mat[pre_idx, :][:, post_idx]).flatten()
                else:
                    new_mat[flat_idx] = self.permute_all(mat[pre_idx, :][:, post_idx]).flatten()

        new_mat = new_mat.reshape(mat.shape)

        assert np.sum(np.isnan(new_mat)) == 0, f"Internal error, not all indexes are given"

        return new_mat


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Export ERODED connection matrix to CSV file")
    parser.add_argument("in_file", help="Snudda HDF5 file with network")
    parser.add_argument("out_file", help="CSV output file")
    parser.add_argument("fraction", help="Fraction of connections kept", type=float)
    parser.add_argument("--full", action="store_false", dest="sparse")
    parser.add_argument("--permute", action="store_true")
    parser.add_argument("--permutation_type",
                        help="Permutation type: 'all', 'types' (preserve types), "
                             "'noself' (dont preserve types, but avoid self connections",
                        default="types")
    args = parser.parse_args()

    s = SnuddaExportErodedConnectionMatrix(args.in_file, args.out_file, save_sparse=args.sparse,
                                           fraction_kept=args.fraction, permute=args.permute,
                                           permute_type=args.permutation_type)



import os
import pickle

import numpy as np
from scipy.spatial import cKDTree

import snudda.utils

# TODO: Move constants like 1000 * sec_x to separate file


class SectionMetaData:

    """ Holds parent_id, children_id, points_id"""

    __slots__ = ["section_id", "parent_section_id", "parent_point_idx", "parent_section_type",
                 "child_section_id", "point_idx", "section_type",
                 "morphology_data", "neuron_id"]

    section_id: int
    parent_section_id: int
    parent_point_idx: int
    parent_section_type: int
    child_section_id: np.ndarray  # First row is child_section_id, second row is child_section_type
    point_idx: np.ndarray
    section_type: int
    morphology_data: object
    neuron_id: int  # Is this set, is it used?

    def __init__(self, section_id, section_type, morphology_data, build_section=True):

        self.morphology_data = morphology_data
        self.section_id = section_id
        self.section_type = section_type

        self.point_idx = None
        self.parent_section_id = None
        self.parent_point_idx = None
        self.parent_section_type = None
        self.child_section_id = None

        if build_section:
            self.build_section()

    def build_section(self):

        idx = np.where((self.morphology_data.section_data[:, 0] == self.section_id)
                       & (self.morphology_data.section_data[:, 2] == self.section_type))[0].astype(np.int32)

        if len(idx) == 0:
            raise ValueError(f"Section id {self.section_id} has no points in morphology_data")

        if not (np.diff(idx) == 1).all():
            raise ValueError(f"Points on section must be consecutive")

        parent_idx = self.morphology_data.section_data[idx[0], 3]
        if parent_idx == -1:
            # Special case, section is soma
            self.point_idx = idx
            self.parent_point_idx = -1
            self.parent_section_type = -1
            self.parent_section_id = -9999
        elif self.morphology_data.section_data[parent_idx, 2] != self.morphology_data.section_data[idx[0], 2]:
            # Special case, root node -- parent section is of different type (e.g. soma -- dend)
            self.point_idx = idx
            self.parent_point_idx = int(self.morphology_data.section_data[self.point_idx[0], 3])
            self.parent_section_type = int(self.morphology_data.section_data[parent_idx, 2])
            self.parent_section_id = int(self.morphology_data.section_data[self.parent_point_idx, 0])
        else:
            self.point_idx = np.concatenate(([self.morphology_data.section_data[idx[0], 3]], idx), dtype=np.int32)
            self.parent_point_idx = int(self.morphology_data.section_data[self.point_idx[0], 3])
            self.parent_section_type = int(self.morphology_data.section_data[parent_idx, 2])
            self.parent_section_id = int(self.morphology_data.section_data[self.parent_point_idx, 0])

        # By definition only the last point in a section can be a parent to other sections
        child_idx = np.where(self.morphology_data.section_data[:, 3] == idx[-1])[0]

        temp_child_section_id = dict()
        for child_section_id, child_type in zip(self.morphology_data.section_data[child_idx, 0],
                                                self.morphology_data.section_data[child_idx, 2]):
            # self.child_section_id is a dict, which holds the child_section_id for different
            # types (e.g. 1=soma, 2=axon, 3=dend, 4=apical)
            if child_type not in temp_child_section_id:
                temp_child_section_id[child_type] = []

            temp_child_section_id[child_type].append(child_section_id)

#        for child_type in self.child_section_id.keys():
#            self.child_section_id[child_type] = np.array(self.child_section_id[child_type])

        section_id = []
        section_type = []
        for child_type in temp_child_section_id.keys():
            section_id += temp_child_section_id[child_type]
            section_type += [child_type]*len(temp_child_section_id[child_type])

        # Convert dictionary to np.array to conserve memory
        self.child_section_id = np.array([section_id, section_type], dtype=np.int16)

        # Also check it above
        bastard_idx = np.where((idx[0] <= self.morphology_data.section_data[:, 3])
                               & (self.morphology_data.section_data[:, 3] < idx[-1]))[0]

        if not (bastard_idx == idx[1:]).all():
            raise ValueError(f"Only last point in section may have children outside section.")

    def section_length(self):

        comp_len = self.morphology_data.geometry[self.point_idx[-1], 4] \
                   - self.morphology_data.geometry[self.point_idx[0], 4]

        # Comment: Soma should not be part of the point_idx, so this is not needed.
        # if self.morphology_data.section_data[self.point_idx[0], 2] == 1:
        #     # If first point (parent point) is soma then we have to be careful and subtract soma radie if inside.
        #     comp_len -= self.morphology_data.geometry[self.point_idx[0], 3]

        if comp_len <= 0:
            raise ValueError(f"Negative section length detected: {self.morphology_data.swc_file}")

        return comp_len

    def clone(self, new_morphology_data, share_memory=True):
        new_smd = SectionMetaData(section_id=self.section_id, section_type=self.section_type,
                                  morphology_data=new_morphology_data, build_section=False)

        if share_memory:
            new_smd.point_idx = self.point_idx
            new_smd.parent_point_idx = self.parent_point_idx
            new_smd.child_section_id = self.child_section_id
            new_smd.parent_section_id = self.parent_section_id

            # Prevent the user from changing these now that the memory is shared
            self.point_idx.setflags(write=False)

            # for cs in self.child_section_id.values():
            #     cs.setflags(write=False)

            self.child_section_id.setflags(write=False)

        else:
            new_smd.point_idx = self.point_idx.copy()
            new_smd.parent_point_idx = self.parent_point_idx
            new_smd.parent_section_id = self.parent_section_id

            new_smd = dict()
            for key, val in self.child_section_id.items():
                new_smd.child_section_id[key] = val.copy()

        return new_smd

    @property
    def position(self):
        return self.morphology_data.geometry[self.point_idx, :3]

    @property
    def radie(self):
        return self.morphology_data.geometry[self.point_idx, 3]

    @property
    def soma_distance(self):
        return self.morphology_data.geometry[self.point_idx, 4]

    @property
    def section_x(self):
        # Double check that this creates a copy of the data before overwriting first element with 0
        sec_x = self.morphology_data.section_data[self.point_idx, 1] / 1e3
        if self.parent_section_type != 1:
            # If parent is not soma, set first section_x to 0
            sec_x[0] = 0
        return sec_x

    def soma_distance_at(self, section_x):
        return np.interp(section_x, self.section_x, self.soma_distance)


class MorphologyData:

    """
        MorphologyData

        This can hold an entire neuron, or a part of a neuron.

        Args:
            swc_file (str): Path to SWC file
                parent_tree_info (tuple, optional): Specify subtree attachment point
                                                (MorphologyData, parent_label, parent_point_idx, arc_factor)

    """

    __slots__ = ["swc_file", "snudda_data", "verbose", "cache_version",
                 "geometry", "section_data", "sections",
                 "point_lookup", "rotation", "position", "parent_tree_info", "is_loaded", "kd_tree_lookup"]

    def __init__(self, swc_file=None, parent_tree_info=None, snudda_data=None,
                 verbose=False, use_cache=True, lazy_loading=False):

        self.swc_file = swc_file
        self.snudda_data = snudda_data
        self.verbose = verbose
        self.cache_version = 1.4

        self.geometry = None      # x, y, z, r, soma_dist (float)
        self.section_data = None  # section_id, section_x (*1000), section_type (int), parent_point_id (int)
        self.sections = None      # dictionary section_id --> SectionMetaData

        self.point_lookup = dict()    # "dend" --> np.array of point_id for dend points

        self.rotation = None
        self.position = None

        self.parent_tree_info = parent_tree_info     # parent tree, if subtree
        self.is_loaded = False
        
        if not lazy_loading and self.swc_file is not None:
            self.load_swc_file(swc_file=self.swc_file, use_cache=use_cache)

        self.kd_tree_lookup = dict()

    def section_iterator_selective(self, section_type, section_id):

        """ Iterates over all sections of a specific type.

        Args:
            section_type: 1 = soma, 2 = axon, 3 = dend
            section_id: ID of sections to iterate over"""

        if section_id is None:
            section_id = self.sections[section_type].keys()

        for sid in section_id:
            yield self.sections[section_type][sid]

    def has_axon(self):
        return len(self.sections[2]) > 0

    def has_dendrite(self):
        return len(self.sections[3]) > 0

    def load_swc_file(self, swc_file=None, remapping_types={4: 3}, use_cache=True):

        """ Loads SWC morphology, not SNUDDA_DATA aware (file must exist).

            Args:
                swc_file (str): Path to swc file
                remapping_types (dict): Remapping of compartment types (default: 4 (apical) -> 3 (normal dendrites))
                use_cache (bool): Save and load neuron morphology to cache
        """
        if swc_file is None:
            swc_file = self.swc_file

        swc_file = snudda.utils.snudda_parse_path(swc_file, self.snudda_data)

        if not os.path.isfile(swc_file):
            raise FileNotFoundError(f"Missing SWC file '{swc_file}'")

        if use_cache:
            cache_file, valid_cache = self.get_cache_file()
            if valid_cache and self.load_cache():
                self.is_loaded = True
                return

        data = np.loadtxt(swc_file)

        if len(data.shape) == 1:
            # This is to handle case when we only have a soma
            data = data.reshape([1, 7])

        if any(np.diff(data[:, 0]) != 1):
            raise IndexError(f"SWC file has gaps in ID numbering ({swc_file})")

        if data[0, 0] != 1:
            raise IndexError(f"ID does not start from 1 ({swc_file})")

        if not (data[0, 2:5] == [0, 0, 0]).all():
            raise ValueError(f"Does not have root centered at origo ({swc_file})")

        if data[0, 6] != -1:
            raise ValueError(f"First element must be root, and have parent ID -1 ({swc_file})")

        if not (data[:, 0] > data[:, 6]).all():
            raise ValueError(f"Parent ID must be lower than row ID ({swc_file})")

        item, count = np.unique(data[:, 0], return_counts=True)
        if (count > 1).any():
            raise ValueError(f"Duplicate index: {item[count > 1]} ({swc_file})")

        if self.parent_tree_info is not None and (data[:, 1] != 2).any():
            raise ValueError(f"Only axonal compartments allowed when subtree of neuron")

        self.geometry = np.zeros((data.shape[0], 5), dtype=np.single)  # 2023-04-24: float -> single, to save memory
        self.geometry[:, :4] = data[:, 2:6] * 1e-6  # x, y, z, r -- converted to meter

        # Store metadata for points
        self.section_data = np.full((data.shape[0], 4), -1, dtype=np.int32)
        self.section_data[:, 2] = data[:, 1]
        self.section_data[0, 3] = -1
        parent_row_id = data[1:, 6].astype(int) - 1
        self.section_data[1:, 3] = parent_row_id

        # This remaps apical dendrites to normal dendrites 4 --> 3 (by default)
        for old_key, new_key in remapping_types.items():
            key_idx = self.section_data[:, 2] == old_key
            self.section_data[key_idx] = new_key

        if (np.abs(self.section_data[:, 2] - data[:, 1]) > 1e-12).any():
            raise ValueError(f"Internal error, non integer ID numbers detected ({swc_file})")

        # We previously removed the points inside the soma, but now keep them
        # Got problems with short dendrites.
        # self.delete_points_inside_soma()

        # OBS, parent_row_id is updated when we delete_points_inside_soma
        parent_row_id = self.section_data[1:, 3]

        # Calculate distance to soma and store in self.geometry
        comp_length = np.linalg.norm(self.geometry[parent_row_id, :3] - self.geometry[1:, :3], axis=1)

        for comp_id, parent_id, c_len in zip(range(1, len(parent_row_id)+1), parent_row_id, comp_length):

            self.geometry[comp_id, 4] = self.geometry[parent_id, 4] + c_len

            # if data[0, 1] == 1 and parent_id == 0:
            #     # We need to subtract soma radius from first compartment connecting to soma
            #     self.geometry[comp_id, 4] = max(0, c_len - self.geometry[0, 3])
            # else:
            #     # distance to soma = parents distance to soma + compartment length
            #     self.geometry[comp_id, 4] = self.geometry[parent_id, 4] + c_len

        if (self.geometry[1:, 4] < 0).any():
            raise ValueError("Found compartments with 0 or negative distance to soma.")

        self.build_tree()

        if use_cache:
            self.save_cache(skip_check=True)  # skip_check since we have not done any rotations

        self.is_loaded = True

    def build_tree(self):

        # New sections are triggered when:
        # -- At branch points
        # -- Change of section type
        parent_id, counts = np.unique(self.section_data[:, 3], return_counts=True)
        branch_id = parent_id[counts > 1]
        type_switch_id = np.where(self.section_data[self.section_data[:, 3], 2] - self.section_data[:, 2] != 0)[0]
        type_switch_id = type_switch_id[type_switch_id != 0]
        edge_id = np.union1d(branch_id, self.section_data[type_switch_id, 3])
        edge_flag = np.zeros((self.section_data.shape[0],), dtype=bool)
        edge_flag[edge_id] = True

        section_counter = dict()

        # Assign section id to all points in section_data
        for idx, row in enumerate(self.section_data):
            section_type = row[2]
            parent_id = row[3]

            if parent_id == -1 or edge_flag[parent_id]:

                # https://github.com/neuronsimulator/nrn/blob/5038de0b79ddf7da9b536639989da4c10dbae7f7/share/lib/hoc/import3d/read_swc.hoc?fbclid=IwAR2kEJOcWkbze8i6G2t9uUVZn5MfmxdSHtm3yzWdP240guJY9KFCalUMvug#L304
                if parent_id == 0 and idx in branch_id:

                    # Special case, parent is soma, and the point itself is a branch point
                    # then mark it as section_type = 0, to not create a one point section
                    self.section_data[idx, 2] = 0
                    section_type = 0

                # Parent point is edge, create new section
                if section_type not in section_counter:
                    section_counter[section_type] = 0
                else:
                    section_counter[section_type] += 1

                section_id = section_counter.get(section_type)
                self.section_data[idx, 0] = section_id
            else:
                # Parent was not an edge, inherit section id
                self.section_data[idx, 0] = self.section_data[parent_id, 0]

        # Calculate section_x for all points in section_data
        for section_type in section_counter:
            for section_id in range(section_counter[section_type]+1):
                idx = np.where((self.section_data[:, 0] == section_id) & (self.section_data[:, 2] == section_type))[0]

                if len(idx) == 1 and self.section_data[idx, 2] == 1:
                    self.section_data[idx, 1] = 1000 * 0.5
                    continue

                if not (np.diff(idx) == 1).all():
                    try:
                        raise ValueError(f"Points on a {section_type} (1=soma, 2=axon, 3=dend) section must be consecutive")
                    except:
                        import traceback
                        print(traceback.format_exc())
                        import pdb
                        pdb.set_trace()

                parent_idx = self.section_data[idx, 3]

                # Use soma distance as a shortcut for calculating length of compartments
                dx = self.geometry[idx, 4] - self.geometry[parent_idx, 4]

                # Special case, if parent is soma, then distance to soma should be set to 0
                if self.section_data[parent_idx[0], 2] == 1:
                    dx[0] = 0

                if len(dx) == 1 and dx[0] == 0:
                    self.section_data[idx, 1] = 0
                else:
                    self.section_data[idx, 1] = 1000 * np.cumsum(dx) / np.sum(dx)

        # Build the actual tree
        self.sections = dict()
        for section_type in section_counter:
            if section_type not in self.sections:
                self.sections[section_type] = dict()

            for section_id in range(0, section_counter[section_type] + 1):
                self.sections[section_type][section_id] = SectionMetaData(section_id=section_id,
                                                                          section_type=section_type,
                                                                          morphology_data=self)
        # Create point lookup
        for section_type in self.sections:
            idx = np.where(self.section_data[:, 2] == section_type)[0]
            self.point_lookup[section_type] = idx

    def save_cache(self, skip_check=False):

        cache_file, _ = self.get_cache_file()

        if not skip_check and (self.rotation is not None or self.position is not None):
            raise ValueError(f"Position and rotation must be None when calling save_cache: {self.swc_file}")

        if self.parent_tree_info is not None:
            raise ValueError(f"Parent tree info must be None when calling save_cache: {self.swc_file}")

        data = dict()
        data["cache_version"] = self.cache_version
        data["swc_file"] = self.swc_file
        data["geometry"] = self.geometry
        data["section_data"] = self.section_data

        data["sections"] = dict()
        for sect_type in self.sections.keys():
            data["sections"][sect_type] = dict()
            for sect_key, sect_value in self.sections[sect_type].items():
                data["sections"][sect_type][sect_key] = dict()
                data["sections"][sect_type][sect_key]["section_id"] = sect_value.section_id
                data["sections"][sect_type][sect_key]["section_type"] = sect_value.section_type

                data["sections"][sect_type][sect_key]["parent_point_idx"] = sect_value.parent_point_idx
                data["sections"][sect_type][sect_key]["parent_section_id"] = sect_value.parent_section_id
                data["sections"][sect_type][sect_key]["parent_section_type"] = sect_value.parent_section_type
                data["sections"][sect_type][sect_key]["point_idx"] = sect_value.point_idx

                # data["sections"][sect_type][sect_key]["child_section_id"] = dict()
                # for ch_key, ch_value in sect_value.child_section_id.items():
                #     data["sections"][sect_type][sect_key]["child_section_id"][ch_key] = ch_value

                data["sections"][sect_type][sect_key]["child_section_id"] = sect_value.child_section_id

        try:
            with open(cache_file, "wb") as f:
                pickle.dump(data, f)
        except:
            print(f"Unable to save cache file {cache_file} -- do you have write permission?")

    def load_cache(self):

        cache_file, valid_cache = self.get_cache_file()
        cache_loaded = False

        if valid_cache:
            try:
                with open(cache_file, "rb") as f:
                    data = pickle.load(f)

                if self.cache_version != data["cache_version"]:
                    raise ValueError(f"Cache version mismatch: {data['cache_version']} (required {self.cache_version})")

                if self.swc_file != data["swc_file"]:
                    raise ValueError(f"Cache mismatch. Different paths:\nRequested: {self.swc_file}\nCached: {data['swc_file']}")

                self.geometry = data["geometry"]
                self.section_data = data["section_data"]

                self.sections = dict()
                for sect_type in data["sections"].keys():
                    assert np.issubdtype(type(sect_type), np.integer), \
                        f"sec_type must be int, found {sect_type} ({type(sect_type)})"
                    self.sections[sect_type] = dict()
                    for sect_id, sect_val in data["sections"][sect_type].items():
                        assert np.issubdtype(type(sect_id), np.integer), \
                            f"sec_id key must be int, found sec_id = {sect_id} ({type(sect_id)}"
                        sec = SectionMetaData(section_id=sect_id, section_type=sect_type,
                                              morphology_data=self, build_section=False)
                        sec.point_idx = sect_val["point_idx"]

                        assert sec.point_idx.dtype == np.int32, f"Old format. New format is 32-bit integer"

                        sec.parent_point_idx = int(sect_val["parent_point_idx"])
                        sec.parent_section_id = int(sect_val["parent_section_id"])

                        sec.parent_section_type = int(sect_val["parent_section_type"])

                        # sec.child_section_id = dict()
                        # for ch_key, ch_val in sect_val["child_section_id"].items():
                        #     sec.child_section_id[ch_key] = ch_val

                        sec.child_section_id = sect_val["child_section_id"]

                        self.sections[sect_type][sect_id] = sec

                cache_loaded = True

            except:
                import traceback
                print(traceback.format_exc())
                print(f"Failed to load cache from {cache_file}")

        return cache_loaded

    def get_cache_file(self):

        swc_file = snudda.utils.snudda_parse_path(self.swc_file, self.snudda_data)
        cache_file = f"{swc_file}-cache.pickle"

        if os.path.isfile(cache_file):
            swc_time = os.path.getmtime(swc_file)
            cache_time = os.path.getmtime(cache_file)

            if cache_time > swc_time:
                # Cache file is newer than swc file
                return cache_file, True

        # No valid cache file found
        return cache_file, False

    def delete_points_inside_soma(self):

        # If no soma, do nothing.
        if self.section_data[0, 2] != 1:
            return

        # We assume soma is in centre
        soma_radius = self.geometry[0, 3]
        dist_to_soma = np.linalg.norm(self.geometry[:, :3], axis=1)
        remove_idx = np.where(dist_to_soma < soma_radius)[0]
        remove_idx = remove_idx[remove_idx > 0]  # Do not remove the soma please

        dont_remove_idx = []
        safety_ctr = 0

        marked_parents = set()  # Checked nodes

        for r_idx in remove_idx[::-1]:
            if r_idx in marked_parents:
                continue

            p_chain = [r_idx]

            p_idx = self.section_data[r_idx, 3]
            while p_idx > 0 and safety_ctr < 1000:
                p_chain.append(p_idx)
                marked_parents.add(p_idx)
                if dist_to_soma[p_idx] > soma_radius:
                    dont_remove_idx += p_chain
                    break

                p_idx = self.section_data[p_idx, 3]
                safety_ctr += 1

        if safety_ctr >= 1000:
            raise ValueError("delete_points_inside_soma: Iteration limit hit, bad swc file?")

        if len(dont_remove_idx) > 0:
            remove_idx = np.array(sorted(list(set(remove_idx) - set(dont_remove_idx))))

            if self.verbose:
                print(f"Found dendrite in {self.swc_file} that goes out, and in and out of soma. "
                      f"Points inside soma {[x+1 for x in remove_idx]}, "
                      f"but these have grandparents outside soma, "
                      f"so keep {[x+1 for x in dont_remove_idx]} (SWC numbering)")

        for r_idx in remove_idx:
            update_parent_idx = np.where(self.section_data[:, 3] == r_idx)[0]
            self.section_data[update_parent_idx, 3] = 0  # Children with removed parents, have soma as stepparent

        # Reindex the parents
        for r_idx in remove_idx[::-1]:
            self.section_data[self.section_data[:, 3] >= r_idx, 3] -= 1

        if len(remove_idx) > 0:

            if self.verbose:
                print(f"Removing {len(remove_idx)} dendrite points inside soma from {self.swc_file}: "
                      f"{[x+1 for x in remove_idx]} (SWC numbering)")

            self.section_data = np.delete(self.section_data, remove_idx, axis=0)
            self.geometry = np.delete(self.geometry, remove_idx, axis=0)

    def shifting_dendrites_onto_soma(self):

        raise DeprecationWarning("This was not how NEURON did it.")

        """ Dendrites should start right at the surface of the soma. Here we delete interior points, and add points
            on surface. Branches which do not touch soma get their innermost point shifted to the soma surface. """

        # If the first point is not a soma, skip this step
        if self.section_data[0, 2] != 1:
            return

        soma_radius = self.geometry[0, 3]

        # We assume soma is at origo, there are checks for that in the load_swc
        dist_to_soma = np.linalg.norm(self.geometry[:, :3], axis=1)
        inside_soma_idx = np.where(dist_to_soma < soma_radius)[0]
        parent_inside_idx = self.section_data[inside_soma_idx, 3]
        remove_idx = np.unique(parent_inside_idx[parent_inside_idx > 0])

        for r_idx in remove_idx[::-1]:
            self.section_data[self.section_data[:, 3] >= r_idx, 3] -= 1

        if len(remove_idx) > 0:
            self.section_data = np.delete(self.section_data, remove_idx, axis=0)
            self.geometry = np.delete(self.geometry, remove_idx, axis=0)

        # New points, with soma as parent, should be shifted to surface of soma
        shift_idx = np.where(self.section_data[:, 3] == 0)[0]

        dx = (self.geometry[shift_idx, :3].T / np.linalg.norm(self.geometry[shift_idx, :3], axis=1)[None, :]).T
        self.geometry[shift_idx, :3] = dx * soma_radius

    def clone(self, position, rotation, parent_tree_info=None, share_memory=True):

        """
            Clone the neuron, and place it.

            Args:
                position (np.array): x,y,z coordinate of cloned neuron
                rotation (np.array): 3x3 rotation matrix
                parent_tree_info (optional): Parent tree
                share_memory (bool): If True some numpy arrays are shared with parent neuron

        """

        if self.position is not None or self.rotation is not None:
            raise ValueError("Not allowed to rotate or position a neuron that has already been rotated or positioned")

        new_md = MorphologyData(verbose=self.verbose)
        new_md.swc_file = self.swc_file

        # We can never share memory for geometry, since repositioning a neuron updates geometry
        new_md.geometry = self.geometry.copy()

        if share_memory:
            # Assuming topology of neuron does not change, these values will be constant
            new_md.section_data = self.section_data
            self.section_data.setflags(write=False)

            for p_key, p_value in self.point_lookup.items():
                new_md.point_lookup[p_key] = p_value
                p_value.setflags(write=False)

        else:
            new_md.section_data = self.section_data.copy()

            for p_key, p_value in self.point_lookup.items():
                new_md.point_lookup[p_key] = p_value.copy()  # !!! Should this be copy.deepcopy?

        new_md.sections = dict()

        for sec_type, sec_type_data in self.sections.items():
            new_md.sections[sec_type] = dict()

            for sec_key, sec_value in sec_type_data.items():
                new_md.sections[sec_type][sec_key] = sec_value.clone(new_morphology_data=new_md,
                                                                     share_memory=share_memory)

        new_md.kd_tree_lookup = dict()
        new_md.parent_tree_info = parent_tree_info

        new_md.place(position=position, rotation=rotation)

        return new_md

    def place(self, position=None, rotation=None, parent_tree_info=None, lazy=False):
        if self.position is not None or self.rotation is not None:
            raise ValueError("Not allowed to rotate or position a neuron that has already been rotated or positioned")

        if isinstance(rotation, (list, tuple)):
            rotation = np.array(rotation)
        else:
            rotation = rotation.copy() if rotation is not None else None

        self.rotation = rotation

        if isinstance(position, (list, tuple)):
            position = np.array(position)
        else:
            position = position.copy() if position is not None else None

        self.position = position

        if lazy:
            return

        if not self.is_loaded:
            self.load_swc_file()

        if parent_tree_info is not None:
            self.parent_tree_info = parent_tree_info  # (MorphologyData, point_idx, arc_factor)

        # Here we assume soma is only a point
        if 1 in self.point_lookup:
            soma_position = self.geometry[self.point_lookup[1], :3]
            if not (soma_position == 0).all():
                raise ValueError("Soma must be centered at origo before placement.")
        elif 2 in self.point_lookup:
            # We have no soma, so it is probably an axonal tree, check that it is centered
            axon_root_position = self.geometry[self.point_lookup[2][0], :3]
            if not (axon_root_position == 0).all():
                raise ValueError("Axon root must be centered at origo before placement.")

        if rotation is not None:
            if not np.abs(np.linalg.det(rotation) - 1) < 1e-10 \
                    or not np.abs(np.matmul(rotation, rotation.T) - np.eye(3) < 1e-10).all():
                raise ValueError(f"Not a valid rotation matrix {rotation}")

            self.geometry[:, :3] = np.matmul(self.rotation, self.geometry[:, :3].T).T

        if self.position is not None:
            self.geometry[:, :3] += self.position

        if self.parent_tree_info is not None:
            # We need to update soma distance for subtree based on distance to parent
            # parent_tree_info = (MorphologyData, point_idx, arc_factor) -- attachment point

            parent_object, parent_point_idx, arc_factor = self.parent_tree_info
            parent_position = parent_object.geometry[parent_point_idx, :3]
            parent_soma_distance = parent_object.geometry[parent_point_idx, 4]

            dist_to_parent = np.linalg.norm(self.position - parent_position)
            self.geometry[:, 4] += parent_soma_distance + dist_to_parent * arc_factor

    def section_iterator(self, section_type=None):

        """ Iterates over all sections of a specific type.

        Args:
            section_type: 1 = soma, 2 = axon, 3 = dend """

        if section_type is None:
            for section_dict in self.sections.values():
                for section in section_dict.values():
                    yield section

        elif section_type in self.sections:
            for section in self.sections[section_type].values():
                yield section

        return

    def get_kd_tree(self, compartment_type):

        if compartment_type not in self.kd_tree_lookup:
            comp_idx = np.where(self.section_data[:, 2] == compartment_type)[0]
            coords = self.geometry[comp_idx, :3]

            self.kd_tree_lookup[compartment_type] = cKDTree(coords)

        return self.kd_tree_lookup[compartment_type]

    def get_closest_point(self, coords, compartment_type):

        """ """

        kd_tree = self.get_kd_tree(compartment_type=compartment_type)
        closest_dist, closest_point_idx = kd_tree.query(coords)
        return closest_dist, closest_point_idx


# Separate method to create a random rotation matrix that distributes rotations evenly (this is non-trivial)
def rand_rotation_matrix(deflection=1.0, rand_nums=None):
    """
    Creates a random rotation matrix.

    deflection: the magnitude of the rotation. For 0, no rotation; for 1, completely random
    rotation. Small deflection => small perturbation.
    rand_nums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
    """

    # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c

    if rand_nums is None:
        rand_nums = np.random.uniform(size=(3,))

    theta, phi, z = rand_nums

    theta = theta * 2.0 * deflection * np.pi  # Rotation about the pole (Z).
    phi = phi * 2.0 * np.pi  # For direction of pole deflection.
    z = z * 2.0 * deflection  # For magnitude of pole deflection.

    # Compute a vector V used for distributing points over the sphere
    # via the reflection I - V Transpose(V).  This formulation of V
    # will guarantee that if x[1] and x[2] are uniformly distributed,
    # the reflected points will be uniform on the sphere.  Note that V
    # has length sqrt(2) to eliminate the 2 in the Householder matrix.

    r = np.sqrt(z)
    vv = (np.sin(phi) * r, np.cos(phi) * r, np.sqrt(2.0 - z))

    st = np.sin(theta)
    ct = np.cos(theta)

    rr = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))

    # Construct the rotation matrix  ( V Transpose(V) - I ) R.
    mm = (np.outer(vv, vv) - np.eye(3)).dot(rr)
    return mm


if __name__ == "__main__":

    # file_name = "/home/hjorth/HBP/BasalGangliaData/data/neurons/striatum/dspn/str-dspn-e150602_c1_D1-mWT-0728MSN01-v20211026/morphology/WT-0728MSN01-cor-rep-ax-res3.swc"
    file_name = "/home/hjorth/HBP/BasalGangliaData/Parkinson/20220225/PD0/neurons/striatum/dspn/26/WT-1215MSN03-cor-rep-ax-res3-var8.swc"
    md = MorphologyData(swc_file=file_name)

    import pdb
    pdb.set_trace()



from typing import List, Any
from snudda.simulate.simulate import SnuddaSimulate
import snudda.neuromodulation.modulation as modulation
import snudda.neuromodulation.translator as translator
from snudda.simulate.nrn_simulator_parallel import NrnSimulatorParallel
from snudda.utils.snudda_path import snudda_parse_path
from snudda.neurons.neuron_model_extended import NeuronModel
from snudda.utils.load import SnuddaLoad
from neuron import h
import json
import numpy as np
import h5py


class SnuddaSimulateNeuromodulationSynapse(SnuddaSimulate):

    """

        Class for simulating neuromodulation using the adaptive mode

    """

    def __init__(self,
                 network_path=None,
                 network_file=None,
                 input_file=None,
                 output_file=None,
                 verbose=False,
                 log_file=None,
                 disable_gap_junctions=True,
                 disable_synapses=False,
                 simulation_config=None,
                 neuromodulator_description=None):
        """

        @type neuromodulation_weight: float

        """
        self.neuromodulation_synapse_ids = None
        self.neuromodulator_description = neuromodulator_description['description']
        self.neuromodulation = dict()
        self.current_cell = None
        self.syn_gpcrs = list()
        self.inplace_gpcrs = list()
        self.cell_modulator = dict()
        self.neuromodulation_weight = neuromodulator_description['weight']
        self.connector = [info['connector'] for info in self.neuromodulator_description.values()]
        self.name_modulation_keys = [c.replace("conc", "") for c in self.connector]
        self.module_connector = [k+'()'for k in self.connector]
        self.mod_str = dict(zip(self.module_connector, self.connector))

        super(SnuddaSimulateNeuromodulationSynapse, self).__init__(network_path=network_path,
                                                                   network_file=network_file,
                                                                   input_file=input_file,
                                                                   output_file=output_file,
                                                                   verbose=verbose,
                                                                   log_file=log_file,
                                                                   disable_gap_junctions=disable_gap_junctions,
                                                                   disable_synapses=disable_synapses,
                                                                   simulation_config=simulation_config)

        # Change the self.custom_setup from None, and execute the custom setup code within this file
        self.cell_type_ion_channels_per_section = self.ion_channels_per_section()
        self.key_list = self.modulation_keys()

        self.write_log(" Using neuromodulation module in Snudda")

    def setup(self):

        self.check_memory_status()
        self.distribute_neurons()
        self.setup_neurons()
        self.check_memory_status()
        self.pc.barrier()

        # Neuromodulation requires this to be run, before connect_network

        # self.synapse_parameters is loaded in self.setup_neurons, hence it is None before

        self.neuromodulation_synapse_ids = [sid for sid, synapse in self.synapse_parameters.items() if
                                            str(synapse[0]) in self.module_connector]
        self.neuromodulation_setup()

        self.connect_network()
        self.check_memory_status()
        self.pc.barrier()

        self.setup_parse_sim_info()

    def neuromodulation_setup(self):

        # This loops through all the synapses, and connects the relevant ones
        # nextRowSet = [ fromRow, toRow ) -- ie range(fromRow,toRow)
        next_row_set = self.find_next_synapse_group(next_row=0)

        while next_row_set is not None:
            # Add the synapses to the neuron
            self.connect_neuron_synapses_gpcr(start_row=next_row_set[0], end_row=next_row_set[1])

            # Find the next group of synapses
            next_row_set = self.find_next_synapse_group(next_row_set[1])  # 2nd number was not included in range

    def get_neuromodulator_synapses(self, synapse_type_ids):

        return [i for i, synapse_type_id in enumerate(synapse_type_ids) if
                synapse_type_id in self.neuromodulation_synapse_ids]

    def modulation_keys(self):

        key_list = list()
        for neuromodulation_key in self.neuromodulator_description.keys():
            key_list.append('level' + neuromodulation_key)
            key_list.append('mod' + neuromodulation_key)
        return key_list

    def ion_channels_per_section(self):

        cell_type = dict()

        for neuromodulator_key, info_for_neuromodulator in self.neuromodulator_description.items():

            for cell_type_name in info_for_neuromodulator['cells'].keys():

                cell_type.update({cell_type_name: dict()})

                for tpart in info_for_neuromodulator['cells'][cell_type_name]['ion_channels'].keys():

                    if tpart in cell_type[cell_type_name].keys():
                        cell_type[cell_type_name][tpart] = cell_type[cell_type_name][tpart] + \
                                                          info_for_neuromodulator['cells'][cell_type_name][
                                                              'ion_channels'][tpart]
                    else:
                        cell_type[cell_type_name].update({tpart: info_for_neuromodulator['cells'][
                            cell_type_name]['ion_channels'][tpart]})

        return cell_type

    def get_ion_channels_per_section(self, cell_type):
        return self.cell_type_ion_channels_per_section[cell_type]

    def connect_neuron_synapses_gpcr(self, start_row, end_row):

        source_id_list, dest_id, dend_sections, sec_id, sec_x, synapse_type_id, axon_distance, \
            conductance, parameter_id = self.get_synapse_info(start_row=start_row, end_row=end_row)

        gpcr_synapse_index = self.get_neuromodulator_synapses(synapse_type_id)

        # Filter away synapse_type_id not conc* connector and check if the start and end row defines the cell,
        # hence you can send all information to add_gpcr_synapse

        channel_modules = [str(self.synapse_parameters[synapse_type_id[i]][0]) for i in gpcr_synapse_index]

        gpcr_synapse_info = np.take([source_id_list, dend_sections, sec_x, sec_id], gpcr_synapse_index, axis=1).transpose()

        # rewrite code as it is sorted on sec_id, jump in step of sec_id and add to dict
        sort_idx = gpcr_synapse_info[:, -1].argsort()

        gpcr_synapse_info = gpcr_synapse_info[sort_idx]

        if gpcr_synapse_info.shape[0] == 0:
            pass
        else:
            self.add_gpcr_synapse(channel_modules, gpcr_synapse_info)

    def add_gpcr_synapse(self, channel_modules, gpcr_info):

        if gpcr_info.shape[0] == 0:
            self.write_log(f'add_gpcr_synapse : Empty gpcr_info for {channel_modules}. Check if the gpcr synapses are correctly defined for all cell types')

        cell = gpcr_info[0][1].cell()
        postcell_type = str(cell).split('_')[0]
        cell_information = dict()
        current_section = gpcr_info[0][-1]
        start_index = 0
        dend_section = gpcr_info[start_index][1]

        for i, dend_info in enumerate(gpcr_info):

            if current_section != dend_info[-1]:

                cell_information.update({dend_section: {'precell': gpcr_info[start_index:i, 0]}})
                cell_information[dend_section].update({'section_dist': gpcr_info[start_index:i, 2]})
                cell_information[dend_section].update({'mod': channel_modules[start_index:i]})

                start_index = i
                dend_section = gpcr_info[start_index][1]
                current_section = gpcr_info[start_index][-1]

        cell_information.update({dend_section: {'precell': gpcr_info[start_index:i, 0]}})
        cell_information[dend_section].update({'section_dist': gpcr_info[start_index:i, 2]})
        cell_information[dend_section].update({'mod': channel_modules[start_index:i]})

        self.connect_gpcr_synapse_to_ion_channels(cell_information, postcell_type)

    def connect_gpcr_synapse_to_ion_channels(self, cell_information, cell_type):

        cell_added_synapses = self.add_gpcrs_in_cell_segments(cell_information=cell_information)

        ion_channels_per_section = self.get_ion_channels_per_section(cell_type)

        for sec, sec_info in cell_information.items():

            tpart = translator.re_translation[sec.name().split('.')[-1].split('[')[0]]

            ion_channels = ion_channels_per_section[tpart]

            for mechanism_name, values in sec.psection()['density_mechs'].items():

                mechanism_name_ptr = mechanism_name + "_ptr"

                if mechanism_name in ion_channels:

                    if self.verbose:
                        for name, value in sec.psection()['density_mechs'][mechanism_name].items():
                            if "maxMod" in name:
                                for v in value:

                                    if self.verbose:
                                        print(f" Value of {name} is {v} for {mechanism_name} on {cell_type}")

                                    assert v > 0, "Modulation value should be positive"

                    level_list = [type_level for type_level in [*sec.psection()['density_mechs'][mechanism_name].keys()] if 'level' in type_level]
                    mod_key_list = [f"mod{n.replace('level','')}_{mechanism_name_ptr}" for n in level_list]
                    sec.insert(mechanism_name_ptr)


                    # Inplace of

                    for syn in self.connector:
                        neuromodulation_key = syn.replace("conc", "")

                        for segment in sec:

                            fake = self.sim.neuron.h.concHld(segment)
                            self.inplace_gpcrs.append(fake)
                            pointer = fake._ref_concentration
                            # talk to NEURON maybe they can help change this, so you don't have to replace the mechanisms, crashes with uninitialised pointers
                            for neurotransmitter_level, mod_key in zip(level_list, mod_key_list):

                                if neuromodulation_key in mod_key:
                                    setattr(segment, mod_key, 1)
                                    self.sim.neuron.h.setpointer(pointer, neurotransmitter_level,
                                                                 getattr(segment, mechanism_name_ptr))

                    for syn in self.connector:

                        neuromodulation_key = syn.replace("conc", "")

                        for segment in sec:
                            seg_x_str = str(segment.x)
                            pointer = cell_added_synapses[sec][seg_x_str][syn]._ref_concentration
                            # talk to NEURON maybe they can help change this, so you don't have to replace the mechanisms, crashes with uninitialised pointers
                            for neurotransmitter_level, mod_key in zip(level_list, mod_key_list):

                                if neuromodulation_key in mod_key:
                                    setattr(segment, mod_key, 1)
                                    self.sim.neuron.h.setpointer(pointer, neurotransmitter_level,
                                                                getattr(segment, mechanism_name_ptr))

                    # Parameterize the pointer version of density_mech, skip level and mod, as that would turn off modulation
                    for param, val in values.items():
                        for i, segmentet in enumerate(sec):
                            if param not in self.key_list:
                                setattr(segmentet, '_'.join([param, mechanism_name_ptr]), val[i])
                                if self.verbose:
                                    print(f"Value {segmentet} {'_'.join([param, mechanism_name_ptr])} : {getattr(segmentet, '_'.join([param, mechanism_name_ptr]))}")
                            else:
                                if self.verbose:
                                    print(f"Value {segmentet} {'_'.join([param, mechanism_name_ptr])} : {getattr(segmentet, '_'.join([param, mechanism_name_ptr]))}")

            for mech_name in ion_channels:
                sec.uninsert(mech_name)

    def add_gpcrs_in_cell_segments(self, cell_information):

        added_synapses = dict()
        # Add all the gpcrs to the segments which have been marked in this cell

        for str_mod_name, syn_in_section in self.mod_str.items():

            for sec, sec_info in cell_information.items():
                step = 1/(sec.nseg*2)
                for seg in sec:

                    synapse_gpcr = getattr(self.sim.neuron.h, syn_in_section)(seg)

                    self.syn_gpcrs.append(synapse_gpcr)

                    str_modules = [str(k) for k in sec_info['mod']]

                    if len(np.where(((seg.x - step) < sec_info['section_dist']) & ((seg.x - step) < sec_info['section_dist']))) > 0:

                        for cell_id_source in np.take(sec_info['precell'], np.where((np.array(str_modules) == str_mod_name) & ((seg.x - step) < sec_info['section_dist']) & ((seg.x - step) < sec_info['section_dist'])))[0]:
                            nc = self.pc.gid_connect(cell_id_source, synapse_gpcr)
                            nc.weight[0] = self.neuromodulation_weight
                            nc.delay = self.synapse_delay
                            nc.threshold = self.spike_threshold

                            #self.net_con_list.append(nc)
                            self.synapse_dict.update({("neuromodulation_nc", 0): nc})
                    self.synapse_dict.update({("neuromodulation", 0): synapse_gpcr})
                    #self.synapse_list.append(syn)

                    if sec in added_synapses.keys() and str(seg.x) in added_synapses[sec].keys():

                        added_synapses[sec][str(seg.x)].update({syn_in_section: synapse_gpcr})

                    elif sec in added_synapses.keys() and str(seg.x) not in added_synapses[sec].keys():

                        added_synapses[sec].update({str(seg.x): {syn_in_section: synapse_gpcr}})
                    else:
                        added_synapses.update({sec: {str(seg.x): {syn_in_section: synapse_gpcr}}})

        return added_synapses

    def get_synapse(self, channel_module, dend_compartment, section_dist):

        cell_name = str(dend_compartment).split("_")[0]

        syn = None
        # add lookup on channel_module to skip split
        if str(channel_module).split('()')[0] not in self.connector:

            channel_module_p = eval(f"self.sim.neuron.h.{str(channel_module).split('()')[0]}_ptr")

            if self.verbose:
                print(dend_compartment(section_dist).point_processes())

            for point_process_in_section in dend_compartment(section_dist).point_processes():

                neuromodulation_name = str(point_process_in_section).split('[')[0]
                neuromodulation_name_key = neuromodulation_name.replace("conc", "")

                if neuromodulation_name in self.connector:

                    syn = channel_module_p(dend_compartment(section_dist))

                    level = [x for x in dir(syn) if 'level' in x]

                    for nkey in level:

                        if nkey.replace("level", "") in self.name_modulation_keys:
                            fake = self.sim.neuron.h.concHld(dend_compartment(section_dist))
                            self.inplace_gpcrs.append(fake)
                            pointer = fake._ref_concentration
                            self.sim.neuron.h.setpointer(pointer, nkey, syn)

                    pointer = point_process_in_section._ref_concentration

                    for neurotransmitter_key in level:
                        # remove this parameter by setting default 1

                        if neuromodulation_name_key in neurotransmitter_key:
                            setattr(syn, f"mod{neurotransmitter_key.replace('level', '')}", 1)

                            modulator = neurotransmitter_key.replace('level', '')
                            receptor_name = str(channel_module).split('()')[0]
                            if modulator in self.neuromodulator_description and cell_name in self.neuromodulator_description[modulator]["cells"]:
                                if receptor_name in self.neuromodulator_description[modulator]["cells"][cell_name]["receptors"]:
                                    parameters = self.neuromodulator_description[modulator]["cells"][cell_name]["receptors"][receptor_name]

                                    for p, v in parameters.items():
                                        setattr(syn, f"{p}{modulator}", v)

                                        if self.verbose:
                                            print(f" Value of {p}{modulator} is {getattr(syn, f'{p}{modulator}')} at {syn} on {cell_name}")
                                            print(f" Value mod{modulator} : {getattr(syn, f'mod{modulator}')}")
                                            assert getattr(syn, f'{p}{modulator}') != 1.0 and getattr(syn, f'{p}{modulator}') > 0, "NeuronModel has not loaded modulation.json," \
                                                                "neuromodulation is not turned on within the model"
                                            assert getattr(syn, f"mod{modulator}") == 1.0

                            self.sim.neuron.h.setpointer(pointer, neurotransmitter_key, syn)

                            if self.verbose:
                                print(f" Value of {neurotransmitter_key} is {getattr(syn, f'{neurotransmitter_key}')} at {syn} on {cell_name}")

        if syn is None:
            syn = channel_module(dend_compartment(section_dist))

        return syn

    def get_external_input_synapse(self, channel_module, section, section_x):

        cell_name = str(section).split("_")[0]
        syn = None

        if self.verbose:
            print(section(section_x).point_processes())

        for point_process_in_section in section(section_x).point_processes():

            if str(point_process_in_section).split('[')[0] in self.connector and str(channel_module).split('()')[0] not in self.connector:

                channel_module_p = eval('self.sim.neuron.h.' + str(channel_module).split('()')[0] + "_ptr")

                syn = channel_module_p(section(section_x))

                neuromodulation_name = str(point_process_in_section).split('[')[0]
                neuromodulation_name_key = neuromodulation_name.replace("conc", "")

                level = [x for x in dir(syn) if 'level' in x]

                for nkey in level:

                    if nkey.replace("level", "") in self.name_modulation_keys:
                        fake = self.sim.neuron.h.concHld(section(section_x))
                        self.inplace_gpcrs.append(fake)
                        pointer = fake._ref_concentration
                        self.sim.neuron.h.setpointer(pointer, nkey, syn)

                pointer = point_process_in_section._ref_concentration

                for neurotransmitter_key in level:

                    if neuromodulation_name_key in neurotransmitter_key:
                        setattr(syn, 'mod' + neurotransmitter_key.replace('level', ''), 1)
                        self.sim.neuron.h.setpointer(pointer, neurotransmitter_key, syn)

                        if self.verbose:
                            print(f" Value of {neurotransmitter_key} is {getattr(syn, f'{neurotransmitter_key}')} at {syn} on {cell_name}")

                        modulator = neurotransmitter_key.replace('level', '')
                        receptor_name = str(channel_module).split('()')[0]
                        if modulator in self.neuromodulator_description and cell_name in self.neuromodulator_description[modulator]["cells"]:
                            if receptor_name in self.neuromodulator_description[modulator]["cells"][cell_name]["receptors"]:
                                parameters = self.neuromodulator_description[modulator]["cells"][cell_name]["receptors"][
                                    receptor_name]

                                for p, v in parameters.items():
                                    setattr(syn, f"{p}{modulator}", v)

                                    if self.verbose:
                                        print(f" Value of {p}{modulator} is {getattr(syn, f'{p}{modulator}')} at {syn} on {cell_name}")
                                        print(f" Value mod{modulator} : {getattr(syn, f'mod{modulator}')}")
                                        assert getattr(syn, f'{p}{modulator}') != 1.0 and getattr(syn,
                                                                                                  f'{p}{modulator}') > 0, "NeuronModel has not loaded modulation.json," \
                                                                                                                          "neuromodulation is not turned on within the model"
                                        assert getattr(syn, f"mod{modulator}") == 1.0

        if syn is None:
            syn = channel_module(section(section_x))

        return syn




import numpy as np
from matplotlib import pyplot as plt
import sys
import os
import json
from collections import OrderedDict


class PlotNodeBenchmark:

    def __init__(self, network_path_list=None):

        self.data = self.merge_data(map(self.load_benchmark, network_path_list))
        self.fig_dir = os.path.join(network_path_list[0], "figures")

    def plot_data(self):

        # stage = ["Init", "Place", "Detect", "Prune"]
        stage = ["Place", "Detect", "Prune"]

        colour = np.array([[55, 126, 184], [77, 175, 74], [152, 78, 163], [255, 127, 0]]) / 255

        nodes = self.data["place"][:, 1]

        duration = np.zeros((len(stage), len(nodes)))

        for idx, s in enumerate(stage):
            duration[idx, :] = self.data[s.lower()][:, 0] / 3600
            assert (self.data[s.lower()][:, 1] == nodes).all(), f"Stage {s} is missing some node values that Place has"

        small_size = 15
        medium_size = 22
        bigger_size = 30

        plt.rc('font', size=small_size)  # controls default text sizes
        plt.rc('axes', titlesize=small_size)  # fontsize of the axes title
        plt.rc('axes', labelsize=medium_size)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=small_size)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=small_size)  # fontsize of the tick labels
        plt.rc('legend', fontsize=small_size)  # legend fontsize
        plt.rc('figure', titlesize=bigger_size)  # fontsize of the figure title

        fig = plt.figure()
        ax = plt.subplot()
        ax.stackplot(nodes, duration, labels=stage)
        ax.legend(loc="upper right")
        ax.set_xlabel("CPU cores (log scale)")
        ax.set_ylabel("Runtime (h)")
        ax.set_xscale("log")
        # ax.set_yscale("log")
        ax.set_xticks(nodes, minor=False)
        ax.set_xticklabels(nodes.astype(int))

        # Prevent cropping of x-label
        fig.subplots_adjust(bottom=0.15)

        if not os.path.isdir(self.fig_dir):
            os.mkdir(self.fig_dir)

        fig_name = os.path.join(self.fig_dir, "benchmark-n-workers.pdf")
        plt.savefig(fig_name, dpi=300)
        plt.show()

    @staticmethod
    def load_benchmark(network_path):

        file_path = os.path.join(network_path, "benchmark_log.json")
        with open(file_path, "r") as f:
            data = json.load(f, object_pairs_hook=OrderedDict)

        new_data = OrderedDict()

        for main_keys in data.keys():

            for item, value in data[main_keys].items():
                new_data[item] = np.array(value)

        return new_data

    @staticmethod
    def merge_data(data_list, sort=True):

        new_data = OrderedDict()

        for data in data_list:
            for item, value in data.items():
                if item not in new_data:
                    new_data[item] = value
                else:
                    new_data[item] = np.concatenate([new_data[item], value])

        if sort:
            # Sort the items in node order
            for item, value in new_data.items():
                num_workers = np.unique(new_data[item][:, 1])
                d = []
                for nw in num_workers:
                    idx = np.where(new_data[item][:, 1] == nw)
                    mean_duration = np.mean(new_data[item][idx, 0])
                    d.append([mean_duration, nw])

                new_data[item] = np.array(d)

        return new_data


if __name__ == "__main__":

    if len(sys.argv) > 1:
        directories = sys.argv[1:]

        pb = PlotNodeBenchmark(network_path_list=directories)
        pb.plot_data()

    else:
        print("You must specify network folder(s) with benchmark data.")